from __future__ import division, with_statement  # confidence high

import glob
import os
import sys

import numpy as np

import pyfits
from pyfits.util import BytesIO
from pyfits.verify import VerifyError
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings, ignore_warnings

from nose.tools import assert_equal, assert_raises, assert_true


class TestHDUListFunctions(PyfitsTestCase):
    def test_update_name(self):
        hdul = pyfits.open(self.data('o4sp040b0_raw.fits'))
        hdul[4].update_ext_name('Jim', "added by Jim")
        hdul[4].update_ext_version(9, "added by Jim")
        assert_equal(hdul[('JIM', 9)].header['extname'], 'JIM')

    def test_hdu_file_bytes(self):
        hdul = pyfits.open(self.data('checksum.fits'))
        res = hdul[0].filebytes()
        assert_equal(res, 11520)
        res = hdul[1].filebytes()
        assert_equal(res, 8640)

    def test_hdulist_file_info(self):
        hdul = pyfits.open(self.data('checksum.fits'))
        res = hdul.fileinfo(0)

        def test_fileinfo(**kwargs):
            assert_equal(res['datSpan'], kwargs.get('datSpan', 2880))
            assert_equal(res['resized'], kwargs.get('resized', False))
            assert_equal(res['filename'], self.data('checksum.fits'))
            assert_equal(res['datLoc'], kwargs.get('datLoc', 8640))
            assert_equal(res['hdrLoc'], kwargs.get('hdrLoc', 0))
            assert_equal(res['filemode'], 'readonly')

        res = hdul.fileinfo(1)
        test_fileinfo(datLoc=17280, hdrLoc=11520)

        hdu = pyfits.ImageHDU(data=hdul[0].data)
        hdul.insert(1, hdu)

        res = hdul.fileinfo(0)
        test_fileinfo(resized=True)

        res = hdul.fileinfo(1)
        test_fileinfo(datSpan=None, resized=True, datLoc=None, hdrLoc=None)

        res = hdul.fileinfo(2)
        test_fileinfo(resized=1, datLoc=17280, hdrLoc=11520)

    def test_create_from_multiple_primary(self):
        """
        Regression test for #145.  Ensure that a validation error occurs when
        saving an HDUList containing multiple PrimaryHDUs.
        """

        hdul = pyfits.HDUList([pyfits.PrimaryHDU(), pyfits.PrimaryHDU()])
        assert_raises(VerifyError, hdul.writeto, self.temp('temp.fits'),
                      output_verify='exception')

    def test_append_primary_to_empty_list(self):
        # Tests appending a Simple PrimaryHDU to an empty HDUList.
        hdul = pyfits.HDUList()
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.append(hdu)
        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', '')]
        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_extension_to_empty_list(self):
        """Tests appending a Simple ImageHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.ImageHDU(np.arange(100, dtype=np.int32))
        hdul.append(hdu)
        info = [(0, 'PRIMARY', 'PrimaryHDU', 4, (100,), 'int32', '')]
        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_table_extension_to_empty_list(self):
        """Tests appending a Simple Table ExtensionHDU to a empty HDUList."""

        hdul = pyfits.HDUList()
        hdul1 = pyfits.open(self.data('tb.fits'))
        hdul.append(hdul1[1])
        info = [(0, 'PRIMARY', 'PrimaryHDU', 4, (), 'uint8', ''),
                (1, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_groupshdu_to_empty_list(self):
        """Tests appending a Simple GroupsHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.GroupsHDU()
        hdul.append(hdu)

        info = [(0, 'PRIMARY', 'GroupsHDU', 8, (), 'uint8',
                 '1 Groups  0 Parameters')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_primary_to_non_empty_list(self):
        """Tests appending a Simple PrimaryHDU to a non-empty HDUList."""

        hdul = pyfits.open(self.data('arange.fits'))
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.append(hdu)

        info = [(0, 'PRIMARY', 'PrimaryHDU', 7, (11, 10, 7), 'int32', ''),
                (1, '', 'ImageHDU', 6, (100,), 'int32', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_extension_to_non_empty_list(self):
        """Tests appending a Simple ExtensionHDU to a non-empty HDUList."""

        hdul = pyfits.open(self.data('tb.fits'))
        hdul.append(hdul[1])

        info = [(0, 'PRIMARY', 'PrimaryHDU', 11, (), 'int16', ''),
                (1, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', ''),
                (2, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-append.fits'))

        assert_equal(pyfits.info(self.temp('test-append.fits'), output=False),
                     info)

    def test_append_groupshdu_to_non_empty_list(self):
        """Tests appending a Simple GroupsHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.append(hdu)
        hdu = pyfits.GroupsHDU()

        assert_raises(ValueError, hdul.append, hdu)

    def test_insert_primary_to_empty_list(self):
        """Tests inserting a Simple PrimaryHDU to an empty HDUList."""
        hdul = pyfits.HDUList()
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.insert(0, hdu)

        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_extension_to_empty_list(self):
        """Tests inserting a Simple ImageHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.ImageHDU(np.arange(100, dtype=np.int32))
        hdul.insert(0, hdu)

        info = [(0, 'PRIMARY', 'PrimaryHDU', 4, (100,), 'int32', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_table_extension_to_empty_list(self):
        """Tests inserting a Simple Table ExtensionHDU to a empty HDUList."""

        hdul = pyfits.HDUList()
        hdul1 = pyfits.open(self.data('tb.fits'))
        hdul.insert(0, hdul1[1])

        info = [(0, 'PRIMARY', 'PrimaryHDU', 4, (), 'uint8', ''),
                (1, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_groupshdu_to_empty_list(self):
        """Tests inserting a Simple GroupsHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.GroupsHDU()
        hdul.insert(0, hdu)

        info = [(0, 'PRIMARY', 'GroupsHDU', 8, (), 'uint8',
                 '1 Groups  0 Parameters')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_primary_to_non_empty_list(self):
        """Tests inserting a Simple PrimaryHDU to a non-empty HDUList."""

        hdul = pyfits.open(self.data('arange.fits'))
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.insert(1, hdu)

        info = [(0, 'PRIMARY', 'PrimaryHDU', 7, (11, 10, 7), 'int32', ''),
                (1, '', 'ImageHDU', 6, (100,), 'int32', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_extension_to_non_empty_list(self):
        """Tests inserting a Simple ExtensionHDU to a non-empty HDUList."""

        hdul = pyfits.open(self.data('tb.fits'))
        hdul.insert(1, hdul[1])

        info = [(0, 'PRIMARY', 'PrimaryHDU', 11, (), 'int16', ''),
                (1, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', ''),
                (2, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_groupshdu_to_non_empty_list(self):
        """Tests inserting a Simple GroupsHDU to an empty HDUList."""

        hdul = pyfits.HDUList()
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul.insert(0, hdu)
        hdu = pyfits.GroupsHDU()

        assert_raises(ValueError, hdul.insert, 1, hdu)

        info = [(0, 'PRIMARY', 'GroupsHDU', 8, (), 'uint8',
                 '1 Groups  0 Parameters'),
                (1, '', 'ImageHDU', 6, (100,), 'int32', '')]

        hdul.insert(0, hdu)

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_groupshdu_to_begin_of_hdulist_with_groupshdu(self):
        """
        Tests inserting a Simple GroupsHDU to the beginning of an HDUList
        that that already contains a GroupsHDU.
        """

        hdul = pyfits.HDUList()
        hdu = pyfits.GroupsHDU()
        hdul.insert(0, hdu)

        assert_raises(ValueError, hdul.insert, 0, hdu)

    def test_insert_extension_to_primary_in_non_empty_list(self):
        # Tests inserting a Simple ExtensionHDU to a non-empty HDUList.
        hdul = pyfits.open(self.data('tb.fits'))
        hdul.insert(0, hdul[1])

        info = [(0, 'PRIMARY', 'PrimaryHDU', 4, (), 'uint8', ''),
                (1, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', ''),
                (2, '', 'ImageHDU', 12, (), 'uint8', ''),
                (3, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_insert_image_extension_to_primary_in_non_empty_list(self):
        """
        Tests inserting a Simple Image ExtensionHDU to a non-empty HDUList
        as the primary HDU.
        """

        hdul = pyfits.open(self.data('tb.fits'))
        hdu = pyfits.ImageHDU(np.arange(100, dtype=np.int32))
        hdul.insert(0, hdu)

        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', ''),
                (1, '', 'ImageHDU', 12, (), 'uint8', ''),
                (2, '', 'BinTableHDU', 24, '2R x 4C', '[1J, 3A, 1E, 1L]', '')]

        assert_equal(hdul.info(output=False), info)

        hdul.writeto(self.temp('test-insert.fits'))

        assert_equal(pyfits.info(self.temp('test-insert.fits'), output=False),
                     info)

    def test_filename(self):
        """Tests the HDUList filename method."""

        hdul = pyfits.open(self.data('tb.fits'))
        name = hdul.filename()
        assert_equal(name, self.data('tb.fits'))

    def test_file_like(self):
        """
        Tests the use of a file like object with no tell or seek methods
        in HDUList.writeto(), HDULIST.flush() or pyfits.writeto()
        """

        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        hdul = pyfits.HDUList()
        hdul.append(hdu)
        tmpfile = open(self.temp('tmpfile.fits'), 'wb')
        hdul.writeto(tmpfile)
        tmpfile.close()

        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', '')]

        assert_equal(pyfits.info(self.temp('tmpfile.fits'), output=False),
                     info)

    def test_file_like_2(self):
        hdu = pyfits.PrimaryHDU(np.arange(100, dtype=np.int32))
        tmpfile = open(self.temp('tmpfile.fits'), 'wb')
        hdul = pyfits.open(tmpfile, mode='ostream')
        hdul.append(hdu)
        hdul.flush()
        tmpfile.close()
        hdul.close()

        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', '')]
        assert_equal(pyfits.info(self.temp('tmpfile.fits'), output=False),
                     info)

    def test_file_like_3(self):

        tmpfile = open(self.temp('tmpfile.fits'), 'wb')
        pyfits.writeto(tmpfile, np.arange(100, dtype=np.int32))
        tmpfile.close()
        info = [(0, 'PRIMARY', 'PrimaryHDU', 5, (100,), 'int32', '')]
        assert_equal(pyfits.info(self.temp('tmpfile.fits'), output=False),
                     info)

    def test_new_hdu_extname(self):
        """
        Tests that new extension HDUs that are added to an HDUList can be
        properly indexed by their EXTNAME/EXTVER (regression test for
        ticket:48).
        """

        f = pyfits.open(self.data('test0.fits'))
        hdul = pyfits.HDUList()
        hdul.append(f[0].copy())
        hdul.append(pyfits.ImageHDU(header=f[1].header))

        assert_equal(hdul[1].header['EXTNAME'], 'SCI')
        assert_equal(hdul[1].header['EXTVER'], 1)
        assert_equal(hdul.index_of(('SCI', 1)), 1)

    def test_update_filelike(self):
        """Test opening a file-like object in update mode and resizing the
        HDU.
        """

        sf = BytesIO()
        arr = np.zeros((100, 100))
        hdu = pyfits.PrimaryHDU(data=arr)
        hdu.writeto(sf)

        sf.seek(0)
        arr = np.zeros((200, 200))
        hdul = pyfits.open(sf, mode='update')
        hdul[0].data = arr
        hdul.flush()

        sf.seek(0)
        hdul = pyfits.open(sf)
        assert_equal(len(hdul), 1)
        assert_true((hdul[0].data == arr).all())

    def test_flush_readonly(self):
        """Test flushing changes to a file opened in a read only mode."""

        oldmtime = os.stat(self.data('test0.fits')).st_mtime
        hdul = pyfits.open(self.data('test0.fits'))
        hdul[0].header['FOO'] = 'BAR'
        with catch_warnings(record=True) as w:
            hdul.flush()
            assert_equal(len(w), 1)
            assert_true('mode is not supported' in str(w[0].message))
        assert_equal(oldmtime, os.stat(self.data('test0.fits')).st_mtime)

    def test_fix_extend_keyword(self):
        hdul = pyfits.HDUList()
        hdul.append(pyfits.PrimaryHDU())
        hdul.append(pyfits.ImageHDU())
        del hdul[0].header['EXTEND']
        hdul.verify('silentfix')

        assert_true('EXTEND' in hdul[0].header)
        assert_equal(hdul[0].header['EXTEND'], True)

    def test_new_hdulist_extend_keyword(self):
        """
        Tests that adding a PrimaryHDU to a new HDUList object updates the
        EXTEND keyword on that HDU.  Regression test for #114.
        """

        h0 = pyfits.Header()
        hdu = pyfits.PrimaryHDU(header=h0)
        sci = pyfits.ImageHDU(data=np.array(10))
        image = pyfits.HDUList([hdu, sci])
        image.writeto(self.temp('temp.fits'))
        assert_true('EXTEND' in hdu.header)
        assert_equal(hdu.header['EXTEND'], True)

    def test_replace_memmaped_array(self):
        # Copy the original before we modify it
        hdul = pyfits.open(self.data('test0.fits'))
        hdul.writeto(self.temp('temp.fits'))

        hdul = pyfits.open(self.temp('temp.fits'), mode='update', memmap=True)
        old_data = hdul[1].data.copy()
        hdul[1].data = hdul[1].data + 1
        hdul.close()
        hdul = pyfits.open(self.temp('temp.fits'), memmap=True)
        assert_true(((old_data + 1) == hdul[1].data).all())

    def test_open_file_with_end_padding(self):
        """Regression test for #106; open files with end padding bytes."""

        hdul = pyfits.open(self.data('test0.fits'),
                           do_not_scale_image_data=True)
        info = hdul.info(output=False)
        hdul.writeto(self.temp('temp.fits'))
        with open(self.temp('temp.fits'), 'ab') as f:
            f.seek(0, os.SEEK_END)
            f.write('\0'.encode('latin1') * 2880)
        with ignore_warnings():
            assert_equal(info,
                         pyfits.info(self.temp('temp.fits'), output=False,
                                     do_not_scale_image_data=True))

    def test_open_file_with_bad_header_padding(self):
        """
        Regression test for #136; open files with nulls for header block
        padding instead of spaces.
        """

        a = np.arange(100).reshape((10, 10))
        hdu = pyfits.PrimaryHDU(data=a)
        hdu.writeto(self.temp('temp.fits'))

        # Figure out where the header padding begins and fill it with nulls
        end_card_pos = str(hdu.header).index('END' + ' ' * 77)
        padding_start = end_card_pos + 80
        padding_len = 2880 - padding_start
        with open(self.temp('temp.fits'), 'r+b') as f:
            f.seek(padding_start)
            f.write('\0'.encode('ascii') * padding_len)

        with catch_warnings(record=True) as w:
            with pyfits.open(self.temp('temp.fits')) as hdul:
                assert_true('contains null bytes instead of spaces' in
                            str(w[0].message))
                assert_equal(len(w), 1)
                assert_equal(len(hdul), 1)
                assert_equal(str(hdul[0].header), str(hdu.header))
                assert_true((hdul[0].data == a).all())

    def test_update_with_truncated_header(self):
        """
        Regression test for #148.  Test that saving an update where the header
        is shorter than the original header doesn't leave a stump from the old
        header in the file.
        """

        data = np.arange(100)
        hdu = pyfits.PrimaryHDU(data=data)
        idx = 1
        while len(hdu.header) < 34:
            hdu.header['TEST%d' % idx] = idx
            idx += 1
        hdu.writeto(self.temp('temp.fits'), checksum=True)

        with pyfits.open(self.temp('temp.fits'), mode='update') as hdul:
            # Modify the header, forcing it to be rewritten
            hdul[0].header['TEST1'] = 2

        with pyfits.open(self.temp('temp.fits')) as hdul:
            assert_true((hdul[0].data == data).all())

    def test_update_resized_header(self):
        """
        Test saving updates to a file where the header is one block smaller
        than before, and in the case where the heade ris one block larger than
        before.
        """

        data = np.arange(100)
        hdu = pyfits.PrimaryHDU(data=data)
        idx = 1
        while len(str(hdu.header)) <= 2880:
            hdu.header['TEST%d' % idx] = idx
            idx += 1
        orig_header = hdu.header.copy()
        hdu.writeto(self.temp('temp.fits'))

        with pyfits.open(self.temp('temp.fits'), mode='update') as hdul:
            while len(str(hdul[0].header)) > 2880:
                del hdul[0].header[-1]

        with pyfits.open(self.temp('temp.fits')) as hdul:
            assert_equal(hdul[0].header, orig_header[:-1])
            assert_true((hdul[0].data == data).all())

        with pyfits.open(self.temp('temp.fits'), mode='update') as hdul:
            idx = 101
            while len(str(hdul[0].header)) <= 2880 * 2:
                hdul[0].header['TEST%d' % idx] = idx
                idx += 1
            # Touch something in the data too so that it has to be rewritten
            hdul[0].data[0] = 27

        with pyfits.open(self.temp('temp.fits')) as hdul:
            assert_equal(hdul[0].header[:-37], orig_header[:-1])
            assert_equal(hdul[0].data[0], 27)
            assert_true((hdul[0].data[1:] == data[1:]).all())

    def test_update_resized_header2(self):
        """
        Regression test for #150.  This is similar to
        test_update_resized_header, but specifically tests a case of multiple
        consecutive flush() calls on the same HDUList object, where each
        flush() requires a resize.
        """

        data1 = np.arange(100)
        data2 = np.arange(100) + 100
        phdu = pyfits.PrimaryHDU(data=data1)
        hdu = pyfits.ImageHDU(data=data2)

        phdu.writeto(self.temp('temp.fits'))

        with pyfits.open(self.temp('temp.fits'), mode='append') as hdul:
            hdul.append(hdu)

        with pyfits.open(self.temp('temp.fits'), mode='update') as hdul:
            idx = 1
            while len(str(hdul[0].header)) <= 2880 * 2:
                hdul[0].header['TEST%d' % idx] = idx
                idx += 1
            hdul.flush()
            hdul.append(hdu)

        with pyfits.open(self.temp('temp.fits')) as hdul:
            assert_true((hdul[0].data == data1).all())
            assert_equal(hdul[1].header, hdu.header)
            assert_true((hdul[1].data == data2).all())
            assert_true((hdul[2].data == data2).all())

    def test_hdul_fromstring(self):
        """
        Test creating the HDUList structure in memory from a string containing
        an entire FITS file.  This is similar to test_hdu_fromstring but for an
        entire multi-extension FITS file at once.
        """

        # Tests HDUList.fromstring for all of PyFITS' built in test files
        def test_fromstring(filename):
            with pyfits.open(self.data(filename)) as hdul:
                orig_info = hdul.info(output=False)
                with open(self.data(filename), 'rb') as f:
                    dat = f.read()

                hdul2 = pyfits.HDUList.fromstring(dat)

                assert_equal(orig_info, hdul2.info(output=False))
                for idx in range(len(hdul)):
                    assert_equal(hdul[idx].header, hdul2[idx].header)
                    if  hdul[idx].data is None or hdul2[idx].data is None:
                        assert_equal(hdul[idx].data, hdul2[idx].data)
                    elif (hdul[idx].data.dtype.fields and
                          hdul2[idx].data.dtype.fields):
                        # Compare tables
                        for n in hdul[idx].data.names:
                            c1 = hdul[idx].data[n]
                            c2 = hdul2[idx].data[n]
                            assert_true((c1 == c2).all())
                    else:
                        assert_true((hdul[idx].data == hdul2[idx].data).all())

        for filename in glob.glob(os.path.join(self.data_dir, '*.fits')):
            if sys.platform == 'win32' and filename == 'zerowidth.fits':
                # Running this test on this file causes a crash in some
                # versions of Numpy on Windows.  See PyFITS ticket #174
                continue
            test_fromstring(os.path.join(self.data_dir, filename))

        # Test that creating an HDUList from something silly raises a TypeError
        assert_raises(TypeError, pyfits.HDUList.fromstring, ['a', 'b', 'c'])

    def test_save_backup(self):
        """Test for #121--save backup of file before flushing changes."""

        self.copy_file('scale.fits')

        with ignore_warnings():
            with pyfits.open(self.temp('scale.fits'), mode='update',
                             save_backup=True) as hdul:
                # Make some changes to the original file to force its header
                # and data to be rewritten
                hdul[0].header['TEST'] = 'TEST'
                hdul[0].data[0] = 0

        assert_true(os.path.exists(self.temp('scale.fits.bak')))
        with pyfits.open(self.data('scale.fits'),
                         do_not_scale_image_data=True) as hdul1:
            with pyfits.open(self.temp('scale.fits.bak'),
                             do_not_scale_image_data=True) as hdul2:
                assert_equal(hdul1[0].header, hdul2[0].header)
                assert_true((hdul1[0].data == hdul2[0].data).all())

        with ignore_warnings():
            with pyfits.open(self.temp('scale.fits'), mode='update',
                             save_backup=True) as hdul:
                # One more time to see if multiple backups are made
                hdul[0].header['TEST2'] = 'TEST'
                hdul[0].data[0] = 1

        assert_true(os.path.exists(self.temp('scale.fits.bak')))
        assert_true(os.path.exists(self.temp('scale.fits.bak.1')))



from __future__ import division


import datetime
import inspect
import os
import re
import warnings

import numpy as np

import pyfits
from pyfits.card import Card
from pyfits.file import _File
from pyfits.header import Header, HEADER_END_RE
from pyfits.util import (lazyproperty, _is_int, _is_pseudo_unsigned,
                         _unsigned_zero, _pad_length, itersubclasses,
                         encode_ascii, decode_ascii, BLOCK_SIZE, deprecated,
                         _get_array_mmap)
from pyfits.verify import _Verify, _ErrList


class _Delayed(object):
    pass
DELAYED = _Delayed()


class InvalidHDUException(Exception):
    """
    A custom exception class used mainly to signal to _BaseHDU.__new__ that
    an HDU cannot possibly be considered valid, and must be assumed to be
    corrupted.
    """


def _hdu_class_from_header(cls, header):
    """
    Used primarily by _BaseHDU.__new__ to find an appropriate HDU class to use
    based on values in the header.  See the _BaseHDU.__new__ docstring.
    """

    klass = cls  # By default, if no subclasses are defined
    if header:
        for c in reversed(list(itersubclasses(cls))):
            try:
                # HDU classes built into pyfits are always considered, but
                # extension HDUs must be explicitly registered
                if not (c.__module__.startswith('pyfits.') or
                        c in cls._hdu_registry):
                    continue
                if c.match_header(header):
                    klass = c
                    break
            except NotImplementedError:
                continue
            except Exception, e:
                warnings.warn(
                    'An exception occurred matching an HDU header to the '
                    'appropriate HDU type: %s' % unicode(e))
                warnings.warn('The HDU will be treated as corrupted.')
                klass = _CorruptedHDU
                break

    return klass


# TODO: Come up with a better __repr__ for HDUs (and for HDULists, for that
# matter)
class _BaseHDU(object):
    """Base class for all HDU (header data unit) classes."""

    _hdu_registry = set()

    # This HDU type is part of the FITS standard
    _standard = True

    # Byte to use for padding out blocks
    _padding_byte = '\x00'

    def __new__(cls, data=None, header=None, *args, **kwargs):
        """
        Iterates through the subclasses of _BaseHDU and uses that class's
        match_header() method to determine which subclass to instantiate.

        It's important to be aware that the class hierarchy is traversed in a
        depth-last order.  Each match_header() should identify an HDU type as
        uniquely as possible.  Abstract types may choose to simply return False
        or raise NotImplementedError to be skipped.

        If any unexpected exceptions are raised while evaluating
        match_header(), the type is taken to be _CorruptedHDU.
        """

        klass = _hdu_class_from_header(cls, header)
        return super(_BaseHDU, cls).__new__(klass)

    def __init__(self, data=None, header=None, *args, **kwargs):
        self._header = header
        self._file = None
        self._buffer = None
        self._hdrLoc = None
        self._datLoc = None
        self._datSpan = None
        self._new = True
        self._output_checksum = False

        if self._header:
            if 'DATASUM' in self._header and 'CHECKSUM' not in self._header:
                self._output_checksum = 'datasum'
            elif 'CHECKSUM' in self._header:
                self._output_checksum = True

        self._name = ''

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, value):
        self._header = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, basestring):
            raise TypeError("'name' attribute must be a string")
        self._name = value

    @property
    def is_image(self):
        return (
            self.name == 'PRIMARY' or
            ('XTENSION' in self._header and
             (self._header['XTENSION'] == 'IMAGE' or
              (self._header['XTENSION'] == 'BINTABLE' and
               'ZIMAGE' in self._header and self._header['ZIMAGE'] == True))))

    @property
    def _data_loaded(self):
        return ('data' in self.__dict__ and self.data is not None and
                self.data is not DELAYED)

    @classmethod
    def register_hdu(cls, hducls):
        cls._hdu_registry.add(hducls)

    @classmethod
    def unregister_hdu(cls, hducls):
        if hducls in cls._hdu_registry:
            cls._hdu_registry.remove(hducls)

    @classmethod
    def match_header(cls, header):
        raise NotImplementedError

    # TODO: This method is a bit kludgy, especially in how it still usually
    # works with a fileobj, and how the first argument may or may not contain
    # HDU data.  This should be rethought.
    @classmethod
    def fromstring(cls, data, fileobj=None, offset=0, checksum=False,
                   ignore_missing_end=False, **kwargs):
        """
        Creates a new HDU object of the appropriate type from a string
        containing the HDU's entire header and, optionally, its data.

        Note: When creating a new HDU from a string without a backing file
        object, the data of that HDU may be read-only.  It depends on whether
        the underlying string was an immutable Python str/bytes object, or some
        kind of read-write memory buffer such as a `memoryview`.

        Parameters
        ----------
        data : str, bytearray, memoryview, ndarray
           A byte string contining the HDU's header and, optionally, its data.
           If `fileobj` is not specified, and the length of `data` extends
           beyond the header, then the trailing data is taken to be the HDU's
           data.  If `fileobj` is specified then the trailing data is ignored.

        fileobj : file (optional)
           The file-like object that this HDU was read from.

        offset : int (optional)
           If `fileobj` is specified, the offset into the file-like object at
           which this HDU begins.

        checksum : bool (optional)
           Check the HDU's checksum and/or datasum.

        ignore_missing_end : bool (optional)
           Ignore a missing end card in the header data.  Note that without
           the end card the end of the header can't be found, so the entire
           data is just assumed to be the header.

        kwargs : (optional)
           May contain additional keyword arguments specific to an HDU type.
           Any unrecognized kwargs are simply ignored.
        """

        if isinstance(data, Header):
            header = data
            if (not len(header) or
                header.keys()[0] not in ('SIMPLE', 'XTENSION')):
                raise ValueError('Block does not begin with SIMPLE or '
                                 'XTENSION')
        else:
            try:
                # Test that the given object supports the buffer interface by
                # ensuring an ndarray can be created from it
                np.ndarray((), dtype='ubyte', buffer=data)
            except TypeError:
                raise TypeError(
                    'The provided object %r does not contain an underlying '
                    'memory buffer.  fromstring() requires an object that '
                    'supports the buffer interface such as bytes, str '
                    '(in Python 2.x but not in 3.x), buffer, memoryview, '
                    'ndarray, etc.' % data)

            if data[:8] not in [encode_ascii('SIMPLE  '),
                                encode_ascii('XTENSION')]:
                raise ValueError('Block does not begin with SIMPLE or '
                                 'XTENSION')

            # Make sure the end card is present
            for match in HEADER_END_RE.finditer(data):
                endpos = match.start()
                if endpos % Card.length == 0:
                    hdrlen = endpos + len(match.group())
                    hdrlen += _pad_length(hdrlen)
                    break
            else:
                if ignore_missing_end:
                    hdrlen = len(data)
                else:
                    raise ValueError('Header missing END card.')

            header = Header.fromstring(decode_ascii(data[:hdrlen]))
        # Determine the appropriate arguments to pass to the constructor from
        # self._kwargs.  self._kwargs contains any number of optional arguments
        # that may or may not be valid depending on the HDU type
        cls = _hdu_class_from_header(cls, header)
        args, varargs, varkwargs, defaults = inspect.getargspec(cls.__init__)
        new_kwargs = kwargs.copy()
        if not varkwargs:
            # If __init__ accepts arbitrary keyword arguments, then we can go
            # ahead and pass all keyword arguments; otherwise we need to delete
            # any that are invalid
            for key in kwargs:
                if key not in args:
                    del new_kwargs[key]

        hdu = cls(data=DELAYED, header=header, **new_kwargs)

        hdu._file = fileobj

        if not fileobj and len(data) > hdrlen:
            # Provide an underlying buffer to read the data from
            hdu._buffer = data

        hdu._hdrLoc = offset                 # beginning of the header area
        if fileobj:
            hdu._datLoc = fileobj.tell()     # beginning of the data area
        else:
            hdu._datLoc = hdrlen

        # data area size, including padding
        size = hdu.size
        hdu._datSpan = size + _pad_length(size)

        # Checksums are not checked on invalid HDU types
        if checksum and checksum != 'remove' and isinstance(hdu, _ValidHDU):
            hdu._verify_checksum_datasum(checksum)

        return hdu

    @classmethod
    def readfrom(cls, fileobj, checksum=False, ignore_missing_end=False,
                 **kwargs):
        """
        Read the HDU from a file.  Normally an HDU should be opened with
        `fitsopen()` which reads the entire HDU list in a FITS file.  But this
        method is still provided for symmetry with `writeto()`.

        Parameters
        ----------
        fileobj : file object or file-like object
            Input FITS file.  The file's seek pointer is assumed to be at the
            beginning of the HDU.

        checksum : bool
            If `True`, verifies that both ``DATASUM`` and
            ``CHECKSUM`` card values (when present in the HDU header)
            match the header and data of all HDU's in the file.

        ignore_missing_end : bool
            Do not issue an exception when opening a file that is
            missing an ``END`` card in the last header.
        """

        # TODO: Figure out a way to make it possible for the _File
        # constructor to be a noop if the argument is already a _File
        if not isinstance(fileobj, _File):
            fileobj = _File(fileobj)

        hdr_offset = fileobj.tell()
        hdr = Header.fromfile(fileobj, endcard=not ignore_missing_end)

        hdu = cls.fromstring(hdr, fileobj=fileobj, offset=hdr_offset,
                             checksum=checksum,
                             ignore_missing_end=ignore_missing_end, **kwargs)

        # If the checksum had to be checked the data may have already been read
        # from the file, in which case we don't want to see relative
        fileobj.seek(hdu._datLoc + hdu._datSpan, os.SEEK_SET)
        return hdu

    def writeto(self, name, output_verify='exception', clobber=False,
                checksum=False):
        """
        Write the HDU to a new file.  This is a convenience method to
        provide a user easier output interface if only one HDU needs
        to be written to a file.

        Parameters
        ----------
        name : file path, file object or file-like object
            Output FITS file.  If opened, must be opened for append
            ("ab+")).

        output_verify : str
            Output verification option.  Must be one of ``"fix"``,
            ``"silentfix"``, ``"ignore"``, ``"warn"``, or
            ``"exception"``.  See :ref:`verify` for more info.

        clobber : bool
            Overwrite the output file if exists.

        checksum : bool
            When `True` adds both ``DATASUM`` and ``CHECKSUM`` cards
            to the header of the HDU when written to the file.
        """

        from pyfits.hdu.hdulist import HDUList

        hdulist = HDUList([self])
        hdulist.writeto(name, output_verify, clobber=clobber,
                        checksum=checksum)

    def _get_raw_data(self, shape, code, offset):
        """
        Return raw array from either the HDU's memory buffer or underlying
        file.
        """

        if isinstance(shape, int):
            shape = (shape,)

        if self._buffer:
            return np.ndarray(shape, dtype=code, buffer=self._buffer,
                              offset=offset)
        else:
            return self._file.readarray(offset=offset, dtype=code, shape=shape)

    # TODO: Rework checksum handling so that it's not necessary to add a
    # checksum argument here
    def _prewriteto(self, checksum=False, inplace=False):
        # If the data is unsigned int 16, 32, or 64 add BSCALE/BZERO
        # cards to header
        if self._data_loaded and self.data is not None and \
           self._standard and _is_pseudo_unsigned(self.data.dtype):
            if 'GCOUNT' in self._header:
                self._header.set('BSCALE', 1, after='GCOUNT')
            else:
                self._header.set('BSCALE', 1)
            self._header.set('BZERO', _unsigned_zero(self.data.dtype),
                             after='BSCALE')

        # Handle checksum

        # If the data is loaded it isn't necessarily 'modified', but we have no
        # way of knowing for sure
        modified = self._header._modified or self._data_loaded

        if checksum == 'remove':
            if 'CHECKSUM' in self._header:
                del self._header['CHECKSUM']

            if 'DATASUM' in self._header:
                del self._header['DATASUM']
        elif modified or self._new:
            if checksum == 'datasum':
                self.add_datasum()
            elif checksum == 'nonstandard_datasum':
                self.add_datasum(blocking='nonstandard')
            elif checksum == 'test':
                self.add_datasum(self._datasum_comment)
                self.add_checksum(self._checksum_comment, True)
            elif checksum == 'nonstandard':
                self.add_checksum(blocking='nonstandard')
            elif checksum:
                self.add_checksum(blocking='standard')

    def _postwriteto(self):
        # If data is unsigned integer 16, 32 or 64, remove the
        # BSCALE/BZERO cards
        if (self._data_loaded and self.data is not None and
            self._standard and _is_pseudo_unsigned(self.data.dtype)):
            for keyword in ('BSCALE', 'BZERO'):
                try:
                    del self._header[keyword]
                except KeyError:
                    pass

    def _writeheader(self, fileobj):
        offset = 0
        if not fileobj.simulateonly:
            try:
                offset = fileobj.tell()
            except (AttributeError, IOError):
                pass

            self._header.tofile(fileobj)

            try:
                size = fileobj.tell() - offset
            except (AttributeError, IOError):
                size = len(str(self._header))
        else:
            size = len(str(self._header))

        # Update hdrLoc with the new offset
        self._hdrLoc = offset

        return offset, size

    def _writedata(self, fileobj):
        # TODO: A lot of the simulateonly stuff should be moved back into the
        # _File class--basically it should turn write and flush into a noop
        offset = 0
        size = 0

        if not fileobj.simulateonly:
            fileobj.flush()
            try:
                offset = fileobj.tell()
            except IOError:
                offset = 0

        if self.data is not None:
            size += self._writedata_internal(fileobj)
            # pad the FITS data block
            if size > 0 and not fileobj.simulateonly:
                padding = _pad_length(size) * self._padding_byte
                # TODO: Not that this is ever likely, but if for some odd
                # reason _padding_byte is > 0x80 this will fail; but really if
                # somebody's custom fits format is doing that, they're doing it
                # wrong and should be reprimanded harshly.
                fileobj.write(padding.encode('ascii'))

        # flush, to make sure the content is written
        if not fileobj.simulateonly:
            fileobj.flush()

        # Update datLoc with the new offset
        self._datLoc = offset
        self._datSpan = size = size + _pad_length(size)

        # return both the location and the size of the data area
        return offset, size

    def _writedata_internal(self, fileobj):
        """
        The beginning and end of most _writedata() implementations are the
        same, but the details of writing the data array itself can vary between
        HDU types, so that should be implemented in this method.

        Should return the size in bytes of the data written.
        """

        if not fileobj.simulateonly:
            fileobj.writearray(self.data)
        return self.data.size * self.data.itemsize

    # TODO: This is the start of moving HDU writing out of the _File class;
    # Though right now this is an internal private method (though still used by
    # HDUList, eventually the plan is to have this be moved into writeto()
    # somehow...
    def _writeto(self, fileobj, inplace=False, copy=False):
        # For now fileobj is assumed to be a _File object
        if not inplace or self._new:
            self._writeheader(fileobj)
            self._writedata(fileobj)
            return

        hdrloc = self._hdrLoc
        hdrsize = self._datLoc - self._hdrLoc
        datloc = self._datLoc
        datsize = self._datSpan

        if self._header._modified:
            # Seek to the original header location in the file
            self._file.seek(hdrloc)
            # This should update hdrloc with he header location in the new file
            hdrloc, hdrsize = self._writeheader(fileobj)

            # If the data is to be written below with self._writedata, that
            # will also properly update the data location; but it should be
            # updated here too
            datloc = hdrloc + hdrsize
        elif copy:
            # Seek to the original header location in the file
            self._file.seek(hdrloc)
            # Before writing, update the hdrloc with the current file position,
            # which is the hdrloc for the new file
            hdrloc = fileobj.tell()
            fileobj.write(self._file.read(hdrsize))
            # The header size is unchanged, but the data location may be
            # different from before depending on if previous HDUs were resized
            datloc = fileobj.tell()
        if self._data_loaded:
            if self.data is not None:
                # Seek through the array's bases for an memmap'd array; we
                # can't rely on the _File object to give us this info since the
                # user may have replaced the previous mmap'd array
                if copy:
                    # Of course, if we're copying the data to a new file we
                    # don't care about flushing the original mmap; instead just
                    # read it into the new file
                    array_mmap = None
                else:
                    array_mmap = _get_array_mmap(self.data)

                if array_mmap is not None:
                    array_mmap.flush()
                else:
                    self._file.seek(self._datLoc)
                    datloc, datsize = self._writedata(fileobj)
        elif copy:
            # Seek to the data location in the original file
            self._file.seek(self._datLoc)
            fileobj.write(self._file.read(datsize))

        self._hdrLoc = hdrloc
        self._datLoc = datloc
        self._datSpan = datsize + _pad_length(datsize)

_AllHDU = _BaseHDU  # For backwards-compatibility, though nobody should have
                    # been using this directly


# For convenience...
# TODO: register_hdu could be made into a class decorator which would be pretty
# cool, but only once 2.6 support is dropped.
register_hdu = _BaseHDU.register_hdu
unregister_hdu = _BaseHDU.unregister_hdu


class _CorruptedHDU(_BaseHDU):
    """
    A Corrupted HDU class.

    This class is used when one or more mandatory `Card`s are
    corrupted (unparsable), such as the ``BITPIX``, ``NAXIS``, or
    ``END`` cards.  A corrupted HDU usually means that the data size
    cannot be calculated or the ``END`` card is not found.  In the case
    of a missing ``END`` card, the `Header` may also contain the binary
    data

    .. note::
       In future, it may be possible to decipher where the last block
       of the `Header` ends, but this task may be difficult when the
       extension is a `TableHDU` containing ASCII data.
    """

    @property
    def size(self):
        """
        Returns the size (in bytes) of the HDU's data part.
        """

        # Note: On compressed files this might report a negative size; but the
        # file is corrupt anyways so I'm not too worried about it.
        if self._buffer is not None:
            return len(self._buffer) - self._datLoc

        return self._file.size - self._datLoc

    def _summary(self):
        return (self.name, 'CorruptedHDU')

    def verify(self):
        pass


class _NonstandardHDU(_BaseHDU, _Verify):
    """
    A Non-standard HDU class.

    This class is used for a Primary HDU when the ``SIMPLE`` Card has
    a value of `False`.  A non-standard HDU comes from a file that
    resembles a FITS file but departs from the standards in some
    significant way.  One example would be files where the numbers are
    in the DEC VAX internal storage format rather than the standard
    FITS most significant byte first.  The header for this HDU should
    be valid.  The data for this HDU is read from the file as a byte
    stream that begins at the first byte after the header ``END`` card
    and continues until the end of the file.
    """

    _standard = False

    @classmethod
    def match_header(cls, header):
        """
        Matches any HDU that has the 'SIMPLE' keyword but is not a standard
        Primary or Groups HDU.
        """

        # The SIMPLE keyword must be in the first card
        card = header.cards[0]

        # The check that 'GROUPS' is missing is a bit redundant, since the
        # match_header for GroupsHDU will always be called before this one.
        if card.keyword == 'SIMPLE':
            if 'GROUPS' not in header and card.value == False:
                return True
            else:
                raise InvalidHDUException
        else:
            return False

    @property
    def size(self):
        """
        Returns the size (in bytes) of the HDU's data part.
        """

        if self._buffer is not None:
            return len(self._buffer) - self._datLoc

        return self._file.size - self._datLoc

    def _writedata(self, fileobj):
        """
        Differs from the base class `_writedata()` in that it doesn't
        automatically add padding, and treats the data as a string of raw bytes
        instead of an array.
        """

        offset = 0
        size = 0

        if not fileobj.simulateonly:
            fileobj.flush()
            try:
                offset = fileobj.tell()
            except IOError:
                offset = 0

        if self.data is not None:
            if not fileobj.simulateonly:
                fileobj.write(self.data)
                # flush, to make sure the content is written
                fileobj.flush()
                size = len(self.data)

        # return both the location and the size of the data area
        return offset, size

    def _summary(self):
        return (self.name, 'NonstandardHDU', len(self._header))

    @lazyproperty
    def data(self):
        """
        Return the file data.
        """

        return self._get_raw_data(self.size, 'ubyte', self._datLoc)

    def _verify(self, option='warn'):
        errs = _ErrList([], unit='Card')

        # verify each card
        for card in self._header.cards:
            errs.append(card._verify(option))

        return errs


class _ValidHDU(_BaseHDU, _Verify):
    """
    Base class for all HDUs which are not corrupted.
    """

    def __init__(self, data=None, header=None, name=None, **kwargs):
        super(_ValidHDU, self).__init__(data=data, header=header)
        if name is not None:
            self.name = name

        if header and not hasattr(self, '_extver'):
            self._extver = header.get('EXTVER', 1)

    @property
    def name(self):
        # Convert the value to a string to be flexible in some pathological
        # cases (see ticket #96)
        if self._header and 'EXTNAME' in self._header:
            self._name = str(self._header['EXTNAME'])
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, basestring):
            raise TypeError("'name' attribute must be a string")
        if not pyfits.EXTENSION_NAME_CASE_SENSITIVE:
            value = value.upper()
        if self._header:
            if 'EXTNAME' in self._header:
                self._header['EXTNAME'] = value
            else:
                self._header.append(('EXTNAME', value, 'extension name'))

    @classmethod
    def match_header(cls, header):
        """
        Matches any HDU that is not recognized as having either the SIMPLE or
        XTENSION keyword in its header's first card, but is nonetheless not
        corrupted.

        TODO: Maybe it would make more sense to use _NonstandardHDU in this
        case?  Not sure...
        """

        return header.keys()[0] not in ('SIMPLE', 'XTENSION')

    @property
    def size(self):
        """
        Size (in bytes) of the data portion of the HDU.
        """

        size = 0
        naxis = self._header.get('NAXIS', 0)
        if naxis > 0:
            size = 1
            for idx in range(naxis):
                size = size * self._header['NAXIS' + str(idx + 1)]
            bitpix = self._header['BITPIX']
            gcount = self._header.get('GCOUNT', 1)
            pcount = self._header.get('PCOUNT', 0)
            size = abs(bitpix) * gcount * (pcount + size) // 8
        return size

    def filebytes(self):
        """
        Calculates and returns the number of bytes that this HDU will write to
        a file.

        Parameters
        ----------
        None

        Returns
        -------
        Number of bytes
        """

        f = _File()
        # TODO: Fix this once new HDU writing API is settled on
        return self._writeheader(f)[1] + self._writedata(f)[1]

    def fileinfo(self):
        """
        Returns a dictionary detailing information about the locations
        of this HDU within any associated file.  The values are only
        valid after a read or write of the associated file with no
        intervening changes to the `HDUList`.

        Parameters
        ----------
        None

        Returns
        -------
        dictionary or None

           The dictionary details information about the locations of
           this HDU within an associated file.  Returns `None` when
           the HDU is not associated with a file.

           Dictionary contents:

           ========== ================================================
           Key        Value
           ========== ================================================
           file       File object associated with the HDU
           filemode   Mode in which the file was opened (readonly, copyonwrite,
                      update, append, ostream)
           hdrLoc     Starting byte location of header in file
           datLoc     Starting byte location of data block in file
           datSpan    Data size including padding
           ========== ================================================
        """

        if hasattr(self, '_file') and self._file:
            return {'file': self._file, 'filemode': self._file.mode,
                    'hdrLoc': self._hdrLoc, 'datLoc': self._datLoc,
                    'datSpan': self._datSpan}
        else:
            return None

    def copy(self):
        """
        Make a copy of the HDU, both header and data are copied.
        """

        if self.data is not None:
            data = self.data.copy()
        else:
            data = None
        return self.__class__(data=data, header=self._header.copy())

    # TODO: self.name should be a property that updates the EXTNAME keyword
    # automatically; likewise for self.version.  Likewise it should
    # automatically get its value from the Header keyword.  This method should
    # just be deprecated, as should update_ext_version
    def update_ext_name(self, value, comment=None, before=None,
                        after=None, savecomment=False):
        """
        Update the extension name associated with the HDU.

        If the keyword already exists in the Header, it's value and/or comment
        will be updated.  If it does not exist, a new card will be created
        and it will be placed before or after the specified location.
        If no `before` or `after` is specified, it will be appended at
        the end.

        Parameters
        ----------
        value : str
            value to be used for the new extension name

        comment : str, optional
            to be used for updating, default=None.

        before : str or int, optional
            name of the keyword, or index of the `Card` before which
            the new card will be placed in the Header.  The argument
            `before` takes precedence over `after` if both specified.

        after : str or int, optional
            name of the keyword, or index of the `Card` after which
            the new card will be placed in the Header.

        savecomment : bool, optional
            When `True`, preserve the current comment for an existing
            keyword.  The argument `savecomment` takes precedence over
            `comment` if both specified.  If `comment` is not
            specified then the current comment will automatically be
            preserved.
        """

        if 'extname' in self._header and savecomment:
            comment = None

        self._header.set('EXTNAME', value, comment, before, after)
        self.name = value

    def update_ext_version(self, value, comment=None, before=None,
                           after=None, savecomment=False):
        """
        Update the extension version associated with the HDU.

        If the keyword already exists in the Header, it's value and/or comment
        will be updated.  If it does not exist, a new card will be created
        and it will be placed before or after the specified location.
        If no `before` or `after` is specified, it will be appended at
        the end.

        Parameters
        ----------
        value : str
            value to be used for the new extension version

        comment : str, optional
            to be used for updating, default=None.

        before : str or int, optional
            name of the keyword, or index of the `Card` before which
            the new card will be placed in the Header.  The argument
            `before` takes precedence over `after` if both specified.

        after : str or int, optional
            name of the keyword, or index of the `Card` after which
            the new card will be placed in the Header.

        savecomment : bool, optional
            When `True`, preserve the current comment for an existing
            keyword.  The argument `savecomment` takes precedence over
            `comment` if both specified.  If `comment` is not
            specified then the current comment will automatically be
            preserved.
        """

        if 'extver' in self._header and savecomment:
            comment = None

        self._header.set('extver', value, comment, before, after)
        self._extver = value

    def _verify(self, option='warn'):
        errs = _ErrList([], unit='Card')

        is_valid = lambda v: v in [8, 16, 32, 64, -32, -64]

        # Verify location and value of mandatory keywords.
        # Do the first card here, instead of in the respective HDU classes, so
        # the checking is in order, in case of required cards in wrong order.
        if isinstance(self, ExtensionHDU):
            firstkey = 'XTENSION'
            firstval = self._extension
        else:
            firstkey = 'SIMPLE'
            firstval = True

        self.req_cards(firstkey, 0, None, firstval, option, errs)
        self.req_cards('BITPIX', 1, lambda v: (_is_int(v) and is_valid(v)), 8,
                       option, errs)
        self.req_cards('NAXIS', 2,
                       lambda v: (_is_int(v) and v >= 0 and v <= 999), 0,
                       option, errs)

        naxis = self._header.get('NAXIS', 0)
        if naxis < 1000:
            for ax in range(3, naxis + 3):
                self.req_cards('NAXIS' + str(ax - 2), ax,
                               lambda v: (_is_int(v) and v >= 0), 1, option,
                               errs)

            # Remove NAXISj cards where j is not in range 1, naxis inclusive.
            for keyword in self._header:
                if keyword.startswith('NAXIS') and len(keyword) > 5:
                    try:
                        number = int(keyword[5:])
                        if number <= 0 or number > naxis:
                            raise ValueError
                    except ValueError:
                        err_text = ("NAXISj keyword out of range ('%s' when "
                                    "NAXIS == %d)" % (keyword, naxis))

                        def fix(self=self, keyword=keyword):
                            del self._header[keyword]

                        errs.append(
                            self.run_option(option=option, err_text=err_text,
                                            fix=fix, fix_text="Deleted."))


        # Verify that the EXTNAME keyword exists and is a string
        if 'EXTNAME' in self._header:
            if not isinstance(self._header['EXTNAME'], basestring):
                err_text = 'The EXTNAME keyword must have a string value.'
                fix_text = 'Converted the EXTNAME keyword to a string value.'

                def fix(header=self._header):
                    header['EXTNAME'] = str(header['EXTNAME'])

                errs.append(self.run_option(option, err_text=err_text,
                                            fix_text=fix_text, fix=fix))

        # verify each card
        for card in self._header.cards:
            errs.append(card._verify(option))

        return errs

    # TODO: Improve this API a little bit--for one, most of these arguments
    # could be optional
    def req_cards(self, keyword, pos, test, fix_value, option, errlist):
        """
        Check the existence, location, and value of a required `Card`.

        TODO: Write about parameters

        If `pos` = `None`, it can be anywhere.  If the card does not exist,
        the new card will have the `fix_value` as its value when created.
        Also check the card's value by using the `test` argument.
        """

        errs = errlist
        fix = None

        try:
            index = self._header.index(keyword)
        except ValueError:
            index = None

        fixable = fix_value is not None

        insert_pos = len(self._header) + 1

        # If pos is an int, insert at the given position (and convert it to a
        # lambda)
        if _is_int(pos):
            insert_pos = pos
            pos = lambda x: x == insert_pos

        # if the card does not exist
        if index is None:
            err_text = "'%s' card does not exist." % keyword
            fix_text = "Fixed by inserting a new '%s' card." % keyword
            if fixable:
                # use repr to accomodate both string and non-string types
                # Boolean is also OK in this constructor
                card = (keyword, fix_value)

                def fix(self=self, insert_pos=insert_pos, card=card):
                    self._header.insert(insert_pos, card)

            errs.append(self.run_option(option, err_text=err_text,
                        fix_text=fix_text, fix=fix, fixable=fixable))
        else:
            # if the supposed location is specified
            if pos is not None:
                if not pos(index):
                    err_text = ("'%s' card at the wrong place (card %d)." %
                                (keyword, index))

                    fix_text = ("Fixed by moving it to the right place "
                                "(card %d)." % insert_pos)

                    def fix(self=self, index=index, insert_pos=insert_pos):
                        card = self._header.cards[index]
                        del self._header[index]
                        self._header.insert(insert_pos, card)

                    errs.append(self.run_option(option, err_text=err_text,
                                fix_text=fix_text, fix=fix))

            # if value checking is specified
            if test:
                val = self._header[keyword]
                if not test(val):
                    err_text = ("'%s' card has invalid value '%s'." %
                                (keyword, val))
                    fix_text = "Fixed by setting a new value '%s'." % fix_value

                    if fixable:
                        def fix(self=self, keyword=keyword, val=fix_value):
                            self._header[keyword] = fix_value

                    errs.append(self.run_option(option, err_text=err_text,
                                fix_text=fix_text, fix=fix, fixable=fixable))

        return errs

    def add_datasum(self, when=None, blocking='standard'):
        """
        Add the ``DATASUM`` card to this HDU with the value set to the
        checksum calculated for the data.

        Parameters
        ----------
        when : str, optional
            Comment string for the card that by default represents the
            time when the checksum was calculated

        blocking: str, optional
            "standard" or "nonstandard", compute sum 2880 bytes at a time, or
            not

        Returns
        -------
        checksum : int
            The calculated datasum

        Notes
        -----
        For testing purposes, provide a `when` argument to enable the
        comment value in the card to remain consistent.  This will
        enable the generation of a ``CHECKSUM`` card with a consistent
        value.
        """

        cs = self._calculate_datasum(blocking)

        if when is None:
            when = 'data unit checksum updated %s' % self._get_timestamp()

        self._header['DATASUM'] = (str(cs), when)
        return cs

    def add_checksum(self, when=None, override_datasum=False,
                     blocking='standard'):
        """
        Add the ``CHECKSUM`` and ``DATASUM`` cards to this HDU with
        the values set to the checksum calculated for the HDU and the
        data respectively.  The addition of the ``DATASUM`` card may
        be overridden.

        Parameters
        ----------
        when : str, optional
           comment string for the cards; by default the comments
           will represent the time when the checksum was calculated

        override_datasum : bool, optional
           add the ``CHECKSUM`` card only

        blocking: str, optional
            "standard" or "nonstandard", compute sum 2880 bytes at a time, or
            not

        Notes
        -----
        For testing purposes, first call `add_datasum` with a `when`
        argument, then call `add_checksum` with a `when` argument and
        `override_datasum` set to `True`.  This will provide
        consistent comments for both cards and enable the generation
        of a ``CHECKSUM`` card with a consistent value.
        """

        if not override_datasum:
            # Calculate and add the data checksum to the header.
            data_cs = self.add_datasum(when, blocking)
        else:
            # Just calculate the data checksum
            data_cs = self._calculate_datasum(blocking)

        if when is None:
            when = 'HDU checksum updated %s' % self._get_timestamp()

        # Add the CHECKSUM card to the header with a value of all zeros.
        if 'DATASUM' in self._header:
            self._header.set('CHECKSUM', '0' * 16, when, before='DATASUM')
        else:
            self._header.set('CHECKSUM', '0' * 16, when)

        self._header['CHECKSUM'] = self._calculate_checksum(data_cs, blocking)

    def verify_datasum(self, blocking='standard'):
        """
        Verify that the value in the ``DATASUM`` keyword matches the value
        calculated for the ``DATASUM`` of the current HDU data.

        blocking: str, optional
            "standard" or "nonstandard", compute sum 2880 bytes at a time, or
            not

        Returns
        -------
        valid : int
           - 0 - failure
           - 1 - success
           - 2 - no ``DATASUM`` keyword present
        """

        if 'DATASUM' in self._header:
            datasum = self._calculate_datasum(blocking)
            if datasum == int(self._header['DATASUM']):
                return 1
            elif blocking == 'either':
                # i.e. standard failed,  try nonstandard
                return self.verify_datasum(blocking='nonstandard')
            else:
                # Failed with all permitted blocking kinds
                return 0
        else:
            return 2

    def verify_checksum(self, blocking='standard'):
        """
        Verify that the value in the ``CHECKSUM`` keyword matches the
        value calculated for the current HDU CHECKSUM.

        blocking: str, optional
            "standard" or "nonstandard", compute sum 2880 bytes at a time, or
            not

        Returns
        -------
        valid : int
           - 0 - failure
           - 1 - success
           - 2 - no ``CHECKSUM`` keyword present
        """

        if 'CHECKSUM' in self._header:
            if 'DATASUM' in self._header:
                datasum = self._calculate_datasum(blocking)
            else:
                datasum = 0
            checksum = self._calculate_checksum(datasum, blocking)
            if checksum == self._header['CHECKSUM']:
                return 1
            elif blocking == 'either':
                # i.e. standard failed,  try nonstandard
                return self.verify_checksum(blocking='nonstandard')
            else:
                # Failed with all permitted blocking kinds
                return 0
        else:
            return 2

    def _verify_checksum_datasum(self, blocking):
        """
        Verify the checksum/datasum values if the cards exist in the header.
        Simply displays warnings if either the checksum or datasum don't match.
        """

        # NOTE:  private data members _checksum and _datasum are
        # used by the utility script "fitscheck" to detect missing
        # checksums.

        if 'CHECKSUM' in self._header:
            self._checksum = self._header['CHECKSUM']
            self._checksum_comment = self._header.comments['CHECKSUM']
            if not self.verify_checksum(blocking):
                warnings.warn('Checksum verification failed for HDU %s.\n' %
                              ((self.name, self._extver),))
            del self._header['CHECKSUM']
        else:
            self._checksum = None
            self._checksum_comment = None

        if 'DATASUM' in self._header:
            self._datasum = self._header['DATASUM']
            self._datasum_comment = self._header.comments['DATASUM']

            if not self.verify_datasum(blocking):
                warnings.warn('Datasum verification failed for HDU %s.\n' %
                              ((self.name, self._extver),))
            del self._header['DATASUM']
        else:
            self._checksum = None
            self._checksum_comment = None
            self._datasum = None
            self._datasum_comment = None

    def _get_timestamp(self):
        """
        Return the current timestamp in ISO 8601 format, with microseconds
        stripped off.

        Ex.: 2007-05-30T19:05:11
        """

        return datetime.datetime.now().isoformat()[:19]

    def _calculate_datasum(self, blocking):
        """
        Calculate the value for the ``DATASUM`` card in the HDU.
        """

        if not self._data_loaded:
            # This is the case where the data has not been read from the file
            # yet.  We find the data in the file, read it, and calculate the
            # datasum.
            if self.size > 0:
                raw_data = self._get_raw_data(self._datSpan, 'ubyte',
                                              self._datLoc)
                return self._compute_checksum(raw_data, blocking=blocking)
            else:
                return 0
        elif self.data is not None:
            return self._compute_checksum(self.data.view('ubyte'),
                                          blocking=blocking)
        else:
            return 0

    def _calculate_checksum(self, datasum, blocking):
        """
        Calculate the value of the ``CHECKSUM`` card in the HDU.
        """

        oldChecksum = self._header['CHECKSUM']
        self._header['CHECKSUM'] = '0' * 16

        # Convert the header to a string.
        s = str(self._header)

        # Calculate the checksum of the Header and data.
        cs = self._compute_checksum(np.fromstring(s, dtype='ubyte'), datasum,
                                    blocking=blocking)

        # Encode the checksum into a string.
        s = self._char_encode(~cs)

        # Return the header card value.
        self._header['CHECKSUM'] = oldChecksum

        return s

    def _compute_checksum(self, bytes, sum32=0, blocking="standard"):
        """
        Compute the ones-complement checksum of a sequence of bytes.

        Parameters
        ----------
        bytes
            a memory region to checksum

        sum32
            incremental checksum value from another region

        blocking
            "standard", "nonstandard", or "either"
            selects the block size on which to perform checksumming,
            originally the blocksize was chosen incorrectly.  "nonstandard"
            selects the original approach,  "standard" selects the
            interoperable blocking size of 2880 bytes.  In the context of
            _compute_checksum, "either" is synonymous with "standard".

        Returns
        -------
        ones complement checksum
        """

        blocklen = {'standard': 2880,
                    'nonstandard': len(bytes),
                    'either': 2880,  # do standard first
                    True: 2880}[blocking]

        sum32 = np.uint32(sum32)
        for i in range(0, len(bytes), blocklen):
            length = min(blocklen, len(bytes) - i)   # ????
            sum32 = self._compute_hdu_checksum(bytes[i:i + length], sum32)
        return sum32

    def _compute_hdu_checksum(self, bytes, sum32=0):
        """
        Translated from FITS Checksum Proposal by Seaman, Pence, and Rots.
        Use uint32 literals as a hedge against type promotion to int64.

        This code should only be called with blocks of 2880 bytes
        Longer blocks result in non-standard checksums with carry overflow
        Historically,  this code *was* called with larger blocks and for that
        reason still needs to be for backward compatibility.
        """

        u8 = np.uint32(8)
        u16 = np.uint32(16)
        uFFFF = np.uint32(0xFFFF)

        if bytes.nbytes % 2:
            last = bytes[-1]
            bytes = bytes[:-1]
        else:
            last = np.uint32(0)

        bytes = bytes.view('>u2')

        hi = sum32 >> u16
        lo = sum32 & uFFFF
        hi += np.add.reduce(bytes[0::2])
        lo += np.add.reduce(bytes[1::2])

        if (bytes.nbytes // 2) % 2:
            lo += last << u8
        else:
            hi += last << u8

        hicarry = hi >> u16
        locarry = lo >> u16

        while hicarry or locarry:
            hi = (hi & uFFFF) + locarry
            lo = (lo & uFFFF) + hicarry
            hicarry = hi >> u16
            locarry = lo >> u16

        return (hi << u16) + lo

    # _MASK and _EXCLUDE used for encoding the checksum value into a character
    # string.
    _MASK = [0xFF000000,
             0x00FF0000,
             0x0000FF00,
             0x000000FF]

    _EXCLUDE = [0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
                0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60]

    def _encode_byte(self, byte):
        """
        Encode a single byte.
        """

        quotient = byte // 4 + ord('0')
        remainder = byte % 4

        ch = np.array(
            [(quotient + remainder), quotient, quotient, quotient],
            dtype='int32')

        check = True
        while check:
            check = False
            for x in self._EXCLUDE:
                for j in [0, 2]:
                    if ch[j] == x or ch[j + 1] == x:
                        ch[j] += 1
                        ch[j + 1] -= 1
                        check = True
        return ch

    def _char_encode(self, value):
        """
        Encodes the checksum `value` using the algorithm described
        in SPR section A.7.2 and returns it as a 16 character string.

        Parameters
        ----------
        value
            a checksum

        Returns
        -------
        ascii encoded checksum
        """

        value = np.uint32(value)

        asc = np.zeros((16,), dtype='byte')
        ascii = np.zeros((16,), dtype='byte')

        for i in range(4):
            byte = (value & self._MASK[i]) >> ((3 - i) * 8)
            ch = self._encode_byte(byte)
            for j in range(4):
                asc[4 * j + i] = ch[j]

        for i in range(16):
            ascii[i] = asc[(i + 15) % 16]

        return decode_ascii(ascii.tostring())


class ExtensionHDU(_ValidHDU):
    """
    An extension HDU class.

    This class is the base class for the `TableHDU`, `ImageHDU`, and
    `BinTableHDU` classes.
    """

    _extension = ''

    @classmethod
    def match_header(cls, header):
        """
        This class should never be instantiated directly.  Either a standard
        extension HDU type should be used for a specific extension, or
        NonstandardExtHDU should be used.
        """

        raise NotImplementedError

    def writeto(self, name, output_verify='exception', clobber=False,
                checksum=False):
        """
        Works similarly to the normal writeto(), but prepends a default
        `PrimaryHDU` are required by extension HDUs (which cannot stand on
        their own).
        """

        from pyfits.hdu.hdulist import HDUList
        from pyfits.hdu.image import PrimaryHDU

        hdulist = HDUList([PrimaryHDU(), self])
        hdulist.writeto(name, output_verify, clobber=clobber,
                        checksum=checksum)

    def _verify(self, option='warn'):

        errs = super(ExtensionHDU, self)._verify(option=option)

        # Verify location and value of mandatory keywords.
        naxis = self._header.get('NAXIS', 0)
        self.req_cards('PCOUNT', naxis + 3, lambda v: (_is_int(v) and v >= 0),
                       0, option, errs)
        self.req_cards('GCOUNT', naxis + 4, lambda v: (_is_int(v) and v == 1),
                       1, option, errs)

        return errs
# For backwards compatilibity, though this needs to be deprecated
# TODO: Mark this as deprecated
_ExtensionHDU = ExtensionHDU


class NonstandardExtHDU(ExtensionHDU):
    """
    A Non-standard Extension HDU class.

    This class is used for an Extension HDU when the ``XTENSION``
    `Card` has a non-standard value.  In this case, pyfits can figure
    out how big the data is but not what it is.  The data for this HDU
    is read from the file as a byte stream that begins at the first
    byte after the header ``END`` card and continues until the
    beginning of the next header or the end of the file.
    """

    _standard = False

    @classmethod
    def match_header(cls, header):
        """
        Matches any extension HDU that is not one of the standard extension HDU
        types.
        """

        card = header.cards[0]
        xtension = card.value
        if isinstance(xtension, basestring):
            xtension = xtension.rstrip()
        # A3DTABLE is not really considered a 'standard' extension, as it was
        # sort of the prototype for BINTABLE; however, since our BINTABLE
        # implementation handles A3DTABLE HDUs it is listed here.
        standard_xtensions = ('IMAGE', 'TABLE', 'BINTABLE', 'A3DTABLE')
        # The check that xtension is not one of the standard types should be
        # redundant.
        return (card.keyword == 'XTENSION' and
                xtension not in standard_xtensions)

    def _summary(self):
        return (self.name, 'NonstandardExtHDU', len(self._header))

    @lazyproperty
    def data(self):
        """
        Return the file data.
        """

        return self._get_raw_data(self.size, 'ubyte', self._datLoc)

# TODO: Mark this as deprecated
_NonstandardExtHDU = NonstandardExtHDU



from __future__ import division # confidence high
from __future__ import with_statement

import numpy as np

import pyfits
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings

from nose.tools import assert_equal


class TestDivisionFunctions(PyfitsTestCase):
    """Test code units that rely on correct integer division."""

    def test_rec_from_string(self):
        t1 = pyfits.open(self.data('tb.fits'))
        s = t1[1].data.tostring()
        a1 = np.rec.array(
                s,
                dtype=np.dtype([('c1', '>i4'), ('c2', '|S3'),
                                ('c3', '>f4'), ('c4', '|i1')]))

    def test_card_with_continue(self):
        h = pyfits.PrimaryHDU()
        with catch_warnings(record=True) as w:
            h.header['abc'] = 'abcdefg' * 20
            assert_equal(len(w), 0)

    def test_valid_hdu_size(self):
        t1 = pyfits.open(self.data('tb.fits'))
        assert_equal(type(t1[1].size), type(1))

    def test_hdu_get_size(self):
        with catch_warnings(record=True) as w:
            t1 = pyfits.open(self.data('tb.fits'))
            assert_equal(len(w), 0)

    def test_section(self):
        # section testing
        fs = pyfits.open(self.data('arange.fits'))
        with catch_warnings(record=True) as w:
            assert_equal(fs[0].section[3, 2, 5], np.array([357]))
            assert_equal(len(w), 0)



from __future__ import with_statement

import os
import time

import numpy as np
from numpy import char as chararray

from nose.tools import assert_equal, assert_true, assert_raises

import pyfits
from pyfits.tests import PyfitsTestCase
from pyfits.tests.test_table import comparerecords


class TestGroupsFunctions(PyfitsTestCase):
    def test_open(self):
        with pyfits.open(self.data('random_groups.fits')) as hdul:
            assert_true(isinstance(hdul[0], pyfits.GroupsHDU))
            naxes = (3, 1, 128, 1, 1)
            parameters = ['UU', 'VV', 'WW', 'BASELINE', 'DATE']
            assert_equal(hdul.info(output=False),
                         [(0, 'PRIMARY', 'GroupsHDU', 147, naxes, 'float32',
                           '3 Groups  5 Parameters')])

            ghdu = hdul[0]
            assert_equal(ghdu.parnames, parameters)
            assert_equal(list(ghdu.data.dtype.names), parameters + ['DATA'])

            assert_true(isinstance(ghdu.data, pyfits.GroupData))
            # The data should be equal to the number of groups
            assert_equal(ghdu.header['GCOUNT'], len(ghdu.data))
            assert_equal(ghdu.data.data.shape, (len(ghdu.data),) + naxes[::-1])
            assert_equal(ghdu.data.parnames, parameters)

            assert_true(isinstance(ghdu.data[0], pyfits.Group))
            assert_equal(len(ghdu.data[0]), len(parameters) + 1)
            assert_equal(ghdu.data[0].data.shape, naxes[::-1])
            assert_equal(ghdu.data[0].parnames, parameters)

    def test_open_groups_in_update_mode(self):
        """
        Test that opening a file containing a groups HDU in update mode and
        then immediately closing it does not result in any unnecessary file
        modifications.

        Similar to
        test_image.TestImageFunctions.test_open_scaled_in_update_mode().
        """

        # Copy the original file before making any possible changes to it
        self.copy_file('random_groups.fits')
        mtime = os.stat(self.temp('random_groups.fits')).st_mtime

        time.sleep(1)

        pyfits.open(self.temp('random_groups.fits'), mode='update',
                    memmap=False).close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert_equal(mtime, os.stat(self.temp('random_groups.fits')).st_mtime)

    def test_parnames_round_trip(self):
        """
        Regression test for #130.  Ensures that opening a random groups file in
        update mode or writing it to a new file does not cause any change to
        the parameter names.
        """

        # Because this test tries to update the random_groups.fits file, let's
        # make a copy of it first (so that the file doesn't actually get
        # modified in the off chance that the test fails
        self.copy_file('random_groups.fits')

        parameters = ['UU', 'VV', 'WW', 'BASELINE', 'DATE']
        with pyfits.open(self.temp('random_groups.fits'), mode='update') as h:
            assert_equal(h[0].parnames, parameters)
            h.flush()
        # Open again just in read-only mode to ensure the parnames didn't
        # change
        with pyfits.open(self.temp('random_groups.fits')) as h:
            assert_equal(h[0].parnames, parameters)
            h.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits')) as h:
            assert_equal(h[0].parnames, parameters)

    def test_groupdata_slice(self):
        """
        A simple test to ensure that slicing GroupData returns a new, smaller
        GroupData object, as is the case with a normal FITS_rec.  This is a
        regression test for an as-of-yet unreported issue where slicing
        GroupData returned a single Group record.
        """


        with pyfits.open(self.data('random_groups.fits')) as hdul:
            s = hdul[0].data[1:]
            assert_true(isinstance(s, pyfits.GroupData))
            assert_equal(len(s), 2)
            assert_equal(hdul[0].data.parnames, s.parnames)

    def test_group_slice(self):
        """
        Tests basic slicing a single group record.
        """

        # A very basic slice test
        with pyfits.open(self.data('random_groups.fits')) as hdul:
            g = hdul[0].data[0]
            s = g[2:4]
            assert_equal(len(s), 2)
            assert_equal(s[0], g[2])
            assert_equal(s[-1], g[-3])
            s = g[::-1]
            assert_equal(len(s), 6)
            assert_true((s[0] == g[-1]).all())
            assert_equal(s[-1], g[0])
            s = g[::2]
            assert_equal(len(s), 3)
            assert_equal(s[0], g[0])
            assert_equal(s[1], g[2])
            assert_equal(s[2], g[4])

    def test_create_groupdata(self):
        """
        Basic test for creating GroupData from scratch.
        """

        imdata = np.arange(100.0)
        imdata.shape = (10, 1, 1, 2, 5)
        pdata1 = np.arange(10, dtype=np.float32) + 0.1
        pdata2 = 42.0
        x = pyfits.hdu.groups.GroupData(imdata, parnames=['abc', 'xyz'],
                                        pardata=[pdata1, pdata2], bitpix=-32)

        assert_equal(x.parnames, ['abc', 'xyz'])
        assert_true((x.par('abc') == pdata1).all())
        assert_true((x.par('xyz') == ([pdata2] * len(x))).all())
        assert_true((x.data == imdata).all())

        # Test putting the data into a GroupsHDU and round-tripping it
        ghdu = pyfits.GroupsHDU(data=x)
        ghdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits')) as h:
            hdr = h[0].header
            assert_equal(hdr['GCOUNT'], 10)
            assert_equal(hdr['PCOUNT'], 2)
            assert_equal(hdr['NAXIS'], 5)
            assert_equal(hdr['NAXIS1'], 0)
            assert_equal(hdr['NAXIS2'], 5)
            assert_equal(hdr['NAXIS3'], 2)
            assert_equal(hdr['NAXIS4'], 1)
            assert_equal(hdr['NAXIS5'], 1)
            assert_equal(h[0].data.parnames, ['abc', 'xyz'])
            assert_true(comparerecords(h[0].data, x))

    def test_duplicate_parameter(self):
        """
        Tests support for multiple parameters of the same name, and ensures
        that the data in duplicate parameters are returned as a single summed
        value.
        """

        imdata = np.arange(100.0)
        imdata.shape = (10, 1, 1, 2, 5)
        pdata1 = np.arange(10, dtype=np.float32) + 1
        pdata2 = 42.0
        x = pyfits.hdu.groups.GroupData(imdata, parnames=['abc', 'xyz', 'abc'],
                                        pardata=[pdata1, pdata2, pdata1],
                                        bitpix=-32)

        assert_equal(x.parnames, ['abc', 'xyz', 'abc'])
        assert_true((x.par('abc') == pdata1 * 2).all())
        assert_equal(x[0].par('abc'), 2)

        # Test setting a parameter
        x[0].setpar(0, 2)
        assert_equal(x[0].par('abc'), 3)
        assert_raises(ValueError, x[0].setpar, 'abc', 2)
        x[0].setpar('abc', (2, 3))
        assert_equal(x[0].par('abc'), 5)
        assert_equal(x.par('abc')[0], 5)
        assert_true((x.par('abc')[1:] == pdata1[1:] * 2).all())

        # Test round-trip
        ghdu = pyfits.GroupsHDU(data=x)
        ghdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits')) as h:
            hdr = h[0].header
            assert_equal(hdr['PCOUNT'], 3)
            assert_equal(hdr['PTYPE1'], 'abc')
            assert_equal(hdr['PTYPE2'], 'xyz')
            assert_equal(hdr['PTYPE3'], 'abc')
            assert_equal(x.parnames, ['abc', 'xyz', 'abc'])
            assert_equal(x.dtype.names, ('abc', 'xyz', '_abc', 'DATA'))
            assert_equal(x.par('abc')[0], 5)
            assert_true((x.par('abc')[1:] == pdata1[1:] * 2).all())



from __future__ import with_statement


import os
import signal
import sys

import nose
from nose.tools import assert_equal, assert_raises

from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings
from pyfits.util import ignore_sigint


class TestUtils(PyfitsTestCase):
    def test_ignore_sigint(self):
        if sys.platform.startswith('win'):
            # Not available in some Python versions on Windows
            raise nose.SkipTest('os.kill() not available')

        @ignore_sigint
        def test():
            with catch_warnings(record=True) as w:
                pid = os.getpid()
                os.kill(pid, signal.SIGINT)
                # One more time, for good measure
                os.kill(pid, signal.SIGINT)
                assert_equal(len(w), 2)
                assert_equal(str(w[0].message),
                             'KeyboardInterrupt ignored until test is '
                             'complete!')

        assert_raises(KeyboardInterrupt, test)



import pyfits

from pyfits.tests import PyfitsTestCase

from nose.tools import assert_equal, assert_true


class TestNonstandardHdus(PyfitsTestCase):
    def test_create_fitshdu(self):
        """
        A round trip test of creating a FitsHDU, adding a FITS file to it,
        writing the FitsHDU out as part of a new FITS file, and then reading
        it and recovering the original FITS file.
        """

        self._test_create_fitshdu(compression=False)

    def test_create_fitshdu_with_compression(self):
        """Same as test_create_fitshdu but with gzip compression enabled."""

        self._test_create_fitshdu(compression=True)

    def _test_create_fitshdu(self, compression=False):
        hdul_orig = pyfits.open(self.data('test0.fits'),
                                do_not_scale_image_data=True)

        fitshdu = pyfits.FitsHDU.fromhdulist(hdul_orig, compress=compression)
        # Just to be meta, let's append to the same hdulist that the fitshdu
        # encapuslates
        hdul_orig.append(fitshdu)
        hdul_orig.writeto(self.temp('tmp.fits'), clobber=True)
        del hdul_orig[-1]

        hdul = pyfits.open(self.temp('tmp.fits'))
        assert_true(isinstance(hdul[-1], pyfits.FitsHDU))

        wrapped = hdul[-1].hdulist
        assert_true(isinstance(wrapped, pyfits.HDUList))

        assert_equal(hdul_orig.info(output=False), wrapped.info(output=False))
        assert_true((hdul[1].data == wrapped[1].data).all())
        assert_true((hdul[2].data == wrapped[2].data).all())
        assert_true((hdul[3].data == wrapped[3].data).all())
        assert_true((hdul[4].data == wrapped[4].data).all())



from __future__ import division
from __future__ import with_statement

import gzip
import os
import sys
import tempfile
import urllib
import warnings
import zipfile

import numpy as np
from numpy import memmap as Memmap

from pyfits.util import (isreadable, iswritable, isfile, fileobj_open,
                         fileobj_name, fileobj_closed, fileobj_mode,
                         _array_from_file, _array_to_file, _write_string,
                         deprecated)


# File object open modes
PYTHON_MODES = {'readonly': 'rb', 'copyonwrite': 'rb', 'update': 'rb+',
                'append': 'ab+', 'ostream': 'wb', 'denywrite': 'rb'}

# readonly actually uses copyonwrite for mmap so that readonly without mmap and
# with mmap still have to same behavior with regard to updating the array.  To
# get a truly readonly mmap use denywrite
# the name 'denywrite' comes from a deprecated flag to mmap() on Linux--it
# should be clarified that 'denywrite' mode is not directly analogous to the
# use of that flag; it was just taken, for lack of anything better, as a name
# that means something like "read only" but isn't readonly.
MEMMAP_MODES = {'readonly': 'c', 'copyonwrite': 'c', 'update': 'r+',
                'append': 'c', 'denywrite': 'r'}

# TODO: Eventually raise a warning, and maybe even later disable the use of
# 'copyonwrite' and 'denywrite' modes unless memmap=True.  For now, however,
# that would generate too many warnings for too many users.  If nothing else,
# wait until the new logging system is in place.

GZIP_MAGIC = u'\x1f\x8b\x08'.encode('raw-unicode-escape')
PKZIP_MAGIC = u'\x50\x4b\x03\x04'.encode('raw-unicode-escape')


class _File(object):
    """
    Represents a FITS file on disk (or in some other file-like object).
    """

    def __init__(self, fileobj=None, mode='readonly', memmap=False):
        if fileobj is None:
            self.__file = None
            self.closed = False
            self.mode = mode
            self.memmap = memmap
            self.compression = None
            self.readonly = False
            self.writeonly = False
            self.simulateonly = True
            return
        else:
            self.simulateonly = False

        if mode not in PYTHON_MODES:
            raise ValueError("Mode '%s' not recognized" % mode)

        if (isinstance(fileobj, basestring) and mode != 'append' and
            not os.path.exists(fileobj) and
            not os.path.splitdrive(fileobj)[0]):
                #
                # Not writing file and file does not exist on local machine and
                # name does not begin with a drive letter (Windows), try to
                # get it over the web.
                #
            try:
                self.name, _ = urllib.urlretrieve(fileobj)
            except (TypeError, ValueError):
                # A couple different exceptions can occur here when passing a
                # filename into urlretrieve in Python 3
                raise IOError('File does not exist: %r' % fileobj)
        else:
            self.name = fileobj_name(fileobj)

        self.closed = False
        self.mode = mode
        self.memmap = memmap

        # Underlying fileobj is a file-like object, but an actual file object
        self.file_like = False

        # More defaults to be adjusted below as necessary
        self.compression = None
        self.readonly = False
        self.writeonly = False

        # Initialize the internal self.__file object
        if isfile(fileobj) or isinstance(fileobj, gzip.GzipFile):
            self._open_fileobj(fileobj, mode)
        elif isinstance(fileobj, basestring):
            self._open_filename(fileobj, mode)
        else:
            self._open_filelike(fileobj, mode)

        if isinstance(fileobj, gzip.GzipFile):
            self.compression = 'gzip'
        elif isinstance(fileobj, zipfile.ZipFile):
            # Reading from zip files is supported but not writing (yet)
            self.compression = 'zip'

        if (mode in ('readonly', 'copyonwrite', 'denywrite') or
                (self.compression and mode == 'update')):
            self.readonly = True
        elif (mode == 'ostream' or
                (self.compression and mode == 'append')):
            self.writeonly = True

        # For 'ab+' mode, the pointer is at the end after the open in
        # Linux, but is at the beginning in Solaris.
        if (mode == 'ostream' or self.compression or
            not hasattr(self.__file, 'seek')):
            # For output stream start with a truncated file.
            # For compressed files we can't really guess at the size
            self.size = 0
        else:
            pos = self.__file.tell()
            self.__file.seek(0, 2)
            self.size = self.__file.tell()
            self.__file.seek(pos)

        if self.memmap and not isfile(self.__file):
            self.memmap = False

    def __repr__(self):
        return '<%s.%s %s>' % (self.__module__, self.__class__.__name__,
                               self.__file)

    # Support the 'with' statement
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @deprecated('3.0', message='This method should not be treated as public.')
    def getfile(self):
        """Will be going away as soon as I figure out how."""

        return self.__file

    def readable(self):
        if self.writeonly:
            return False
        return isreadable(self.__file)

    def read(self, size=None):
        if not hasattr(self.__file, 'read'):
            raise EOFError
        return self.__file.read(size)

    def readarray(self, size=None, offset=0, dtype=np.uint8, shape=None):
        """
        Similar to file.read(), but returns the contents of the underlying
        file as a numpy array (or mmap'd array if memmap=True) rather than a
        string.

        Usually it's best not to use the `size` argument with this method, but
        it's provided for compatibility.
        """

        if not hasattr(self.__file, 'read'):
            raise EOFError

        if not isinstance(dtype, np.dtype):
            dtype = np.dtype(dtype)

        if size and size % dtype.itemsize != 0:
            raise ValueError('size %d not a multiple of %s' % (size, dtype))

        if isinstance(shape, int):
            shape = (shape,)

        if size and shape:
            actualsize = sum(dim * dtype.itemsize for dim in shape)
            if actualsize < size:
                raise ValueError('size %d is too few bytes for a %s array of '
                                 '%s' % (size, shape, dtype))
            if actualsize < size:
                raise ValueError('size %d is too many bytes for a %s array of '
                                 '%s' % (size, shape, dtype))

        if size and not shape:
            shape = (size // dtype.itemsize,)

        if not (size or shape):
            warnings.warn('No size or shape given to readarray(); assuming a '
                          'shape of (1,)')
            shape = (1,)

        if self.memmap:
            return Memmap(self.__file, offset=offset,
                          mode=MEMMAP_MODES[self.mode], dtype=dtype,
                          shape=shape).view(np.ndarray)
        else:
            count = reduce(lambda x, y: x * y, shape)
            pos = self.__file.tell()
            self.__file.seek(offset)
            data = _array_from_file(self.__file, dtype, count, '')
            data.shape = shape
            self.__file.seek(pos)
            return data

    def writable(self):
        if self.readonly:
            return False
        return iswritable(self.__file)

    def write(self, string):
        if hasattr(self.__file, 'write'):
            _write_string(self.__file, string)

    def writearray(self, array):
        """
        Similar to file.write(), but writes a numpy array instead of a string.

        Also like file.write(), a flush() or close() may be needed before
        the file on disk reflects the data written.
        """

        if hasattr(self.__file, 'write'):
            _array_to_file(array, self.__file)

    def flush(self):
        if hasattr(self.__file, 'flush'):
            self.__file.flush()

    def seek(self, offset, whence=0):
        # In newer Python versions, GzipFiles support the whence argument, but
        # I don't think it was added until 2.6; instead of assuming it's
        # present, we implement our own support for it here
        if not hasattr(self.__file, 'seek'):
            return
        if isinstance(self.__file, gzip.GzipFile):
            if whence:
                if whence == 1:
                    offset = self.__file.offset + offset
                else:
                    raise ValueError('Seek from end not supported')
            self.__file.seek(offset)
        else:
            self.__file.seek(offset, whence)

        pos = self.__file.tell()
        if self.size and pos > self.size:
            warnings.warn('File may have been truncated: actual file length '
                          '(%i) is smaller than the expected size (%i)' %
                          (self.size, pos))

    def tell(self):
        if not hasattr(self.__file, 'tell'):
            raise EOFError
        return self.__file.tell()

    def truncate(self, size=None):
        if hasattr(self.__file, 'truncate'):
            self.__file.truncate(size)

    def close(self):
        """
        Close the 'physical' FITS file.
        """

        if hasattr(self.__file, 'close'):
            self.__file.close()

        self.closed = True

    def _open_fileobj(self, fileobj, mode):
        """Open a FITS file from a file object or a GzipFile object."""

        closed = fileobj_closed(fileobj)
        fmode = fileobj_mode(fileobj) or PYTHON_MODES[mode]

        if not closed:
            # In some cases (like on Python 3) a file opened for appending
            # still shows a mode of 'r+', hence the extra check for the append
            # case
            if ((mode == 'append' and fmode not in ('ab+', 'rb+')) or
                (mode != 'append' and PYTHON_MODES[mode] != fmode)):
                raise ValueError(
                    "Input mode '%s' (%s) does not match mode of the "
                    "input file (%s)." % (mode, PYTHON_MODES[mode], fmode))
            self.__file = fileobj
        elif isfile(fileobj):
            self.__file = fileobj_open(self.name, PYTHON_MODES[mode])
            # Return to the beginning of the file--in Python 3 when opening in
            # append mode the file pointer is at the end of the file
            self.__file.seek(0)
        else:
            self.__file = gzip.open(self.name, PYTHON_MODES[mode])

    def _open_filelike(self, fileobj, mode):
        """Open a FITS file from a file-like object, i.e. one that has
        read and/or write methods.
        """

        self.file_like = True
        self.__file = fileobj

        # If there is not seek or tell methods then set the mode to
        # output streaming.
        if (not hasattr(self.__file, 'seek') or
            not hasattr(self.__file, 'tell')):
            self.mode = mode = 'ostream'

        if (self.mode in ('copyonwrite', 'update', 'append') and
            not hasattr(self.__file, 'write')):
            raise IOError("File-like object does not have a 'write' "
                          "method, required for mode '%s'."
                          % self.mode)

        if (self.mode in ('readonly', 'denywrite') and
            not hasattr(self.__file, 'read')):
            raise IOError("File-like object does not have a 'read' "
                          "method, required for mode %r."
                          % self.mode)

    def _open_filename(self, filename, mode):
        """Open a FITS file from a filename string."""

        if os.path.exists(self.name):
            with fileobj_open(self.name, 'rb') as f:
                magic = f.read(4)
        else:
            magic = ''.encode('raw-unicode-escape')
        ext = os.path.splitext(self.name)[1]
        if ext == '.gz' or magic.startswith(GZIP_MAGIC):
            # Handle gzip files
            self.__file = gzip.open(self.name, PYTHON_MODES[mode])
            self.compression = 'gzip'
        elif ext == '.zip' or magic.startswith(PKZIP_MAGIC):
            # Handle zip files
            if mode in ['update', 'append']:
                raise IOError(
                      "Writing to zipped fits files is not currently "
                      "supported")
            zfile = zipfile.ZipFile(self.name)
            namelist = zfile.namelist()
            if len(namelist) != 1:
                raise IOError(
                  "Zip files with multiple members are not supported.")
            self.__file = tempfile.NamedTemporaryFile(suffix='.fits')
            self.__file.write(zfile.read(namelist[0]))
            zfile.close()
            self.compression = 'zip'
        else:
            self.__file = fileobj_open(self.name, PYTHON_MODES[mode])
            # Make certain we're back at the beginning of the file
        self.__file.seek(0)



from __future__ import division  # confidence high
from __future__ import with_statement

import math
import os
import time
import warnings

import numpy as np

import pyfits
from pyfits.tests import PyfitsTestCase
from pyfits.tests.util import catch_warnings, ignore_warnings, CaptureStdio

from nose.tools import (assert_equal, assert_raises, assert_true, assert_false,
                        assert_not_equal)
from numpy.testing import assert_allclose


class TestImageFunctions(PyfitsTestCase):
    def test_constructor_name_arg(self):
        """Like the test of the same name in test_table.py"""

        hdu = pyfits.ImageHDU()
        assert_equal(hdu.name, '')
        assert_true('EXTNAME' not in hdu.header)
        hdu.name = 'FOO'
        assert_equal(hdu.name, 'FOO')
        assert_equal(hdu.header['EXTNAME'], 'FOO')

        # Passing name to constructor
        hdu = pyfits.ImageHDU(name='FOO')
        assert_equal(hdu.name, 'FOO')
        assert_equal(hdu.header['EXTNAME'], 'FOO')

        # And overriding a header with a different extname
        hdr = pyfits.Header()
        hdr['EXTNAME'] = 'EVENTS'
        hdu = pyfits.ImageHDU(header=hdr, name='FOO')
        assert_equal(hdu.name, 'FOO')
        assert_equal(hdu.header['EXTNAME'], 'FOO')

    def test_constructor_copies_header(self):
       """
       Regression test for #153.  Ensure that a header from one HDU is copied
       when used to initialize new HDU.
       """

       ifd = pyfits.HDUList(pyfits.PrimaryHDU())
       phdr = ifd[0].header
       phdr['FILENAME'] = 'labq01i3q_rawtag.fits'

       primary_hdu = pyfits.PrimaryHDU(header=phdr)
       ofd = pyfits.HDUList(primary_hdu)
       ofd[0].header['FILENAME'] = 'labq01i3q_flt.fits'

       # Original header should be unchanged
       assert_equal(phdr['FILENAME'], 'labq01i3q_rawtag.fits')

    def test_open(self):
        # The function "open" reads a FITS file into an HDUList object.  There
        # are three modes to open: "readonly" (the default), "append", and
        # "update".

        # Open a file read-only (the default mode), the content of the FITS
        # file are read into memory.
        r = pyfits.open(self.data('test0.fits')) # readonly

        # data parts are latent instantiation, so if we close the HDUList
        # without touching data, data can not be accessed.
        r.close()
        assert_raises(Exception, lambda x: x[1].data[:2,:2], r)

    def test_open_2(self):
        r = pyfits.open(self.data('test0.fits'))

        info = ([(0, 'PRIMARY', 'PrimaryHDU', 138, (), 'int16', '')] +
                [(x, 'SCI', 'ImageHDU', 61, (40, 40), 'int16', '')
                 for x in range(1, 5)])

        try:
            assert_equal(r.info(output=False), info)
        finally:
            r.close()

    def test_primary_with_extname(self):
        """Regression test for #151.

        Tests that the EXTNAME keyword works with Primary HDUs as well, and
        interacts properly with the .name attribute.  For convenience
        hdulist['PRIMARY'] will still refer to the first HDU even if it has an
        EXTNAME not equal to 'PRIMARY'.
        """

        prihdr = pyfits.Header([('EXTNAME', 'XPRIMARY'), ('EXTVER', 1)])
        hdul = pyfits.HDUList([pyfits.PrimaryHDU(header=prihdr)])
        assert_true('EXTNAME' in hdul[0].header)
        assert_equal(hdul[0].name, 'XPRIMARY')
        assert_equal(hdul[0].name, hdul[0].header['EXTNAME'])

        info = [(0, 'XPRIMARY', 'PrimaryHDU', 5, (), 'uint8', '')]
        assert_equal(hdul.info(output=False), info)

        assert_true(hdul['PRIMARY'] is hdul['XPRIMARY'])
        assert_true(hdul['PRIMARY'] is hdul[('XPRIMARY', 1)])

        hdul[0].name = 'XPRIMARY2'
        assert_equal(hdul[0].header['EXTNAME'], 'XPRIMARY2')

        hdul.writeto(self.temp('test.fits'))
        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_equal(hdul[0].name, 'XPRIMARY2')

    def test_io_manipulation(self):
        # Get a keyword value.  An extension can be referred by name or by
        # number.  Both extension and keyword names are case insensitive.
        with pyfits.open(self.data('test0.fits')) as r:
            assert_equal(r['primary'].header['naxis'], 0)
            assert_equal(r[0].header['naxis'], 0)

            # If there are more than one extension with the same EXTNAME value,
            # the EXTVER can be used (as the second argument) to distinguish
            # the extension.
            assert_equal(r['sci',1].header['detector'], 1)

            # append (using "update()") a new card
            r[0].header['xxx'] = 1.234e56

            assert_equal(str(r[0].header.ascard[-3:]),
                "EXPFLAG = 'NORMAL            ' / Exposure interruption indicator                \n"
                "FILENAME= 'vtest3.fits'        / File name                                      \n"
                "XXX     =            1.234E+56                                                  ")

            # rename a keyword
            r[0].header.rename_key('filename', 'fname')
            assert_raises(ValueError, r[0].header.rename_key, 'fname',
                          'history')

            assert_raises(ValueError, r[0].header.rename_key, 'fname',
                          'simple')
            r[0].header.rename_key('fname', 'filename')

            # get a subsection of data
            assert_true((r[2].data[:3,:3] ==
                         np.array([[349, 349, 348],
                                   [349, 349, 347],
                                   [347, 350, 349]], dtype=np.int16)).all())

            # We can create a new FITS file by opening a new file with "append"
            # mode.
            with pyfits.open(self.temp('test_new.fits'), mode='append') as n:
                # Append the primary header and the 2nd extension to the new
                # file.
                n.append(r[0])
                n.append(r[2])

                # The flush method will write the current HDUList object back
                # to the newly created file on disk.  The HDUList is still open
                # and can be further operated.
                n.flush()
                assert_equal(n[1].data[1,1], 349)

                # modify a data point
                n[1].data[1,1] = 99

                # When the file is closed, the most recent additions of
                # extension(s) since last flush() will be appended, but any HDU
                # already existed at the last flush will not be modified
            del n

            # If an existing file is opened with "append" mode, like the
            # readonly mode, the HDU's will be read into the HDUList which can
            # be modified in memory but can not be written back to the original
            # file.  A file opened with append mode can only add new HDU's.
            os.rename(self.temp('test_new.fits'),
                      self.temp('test_append.fits'))

            with pyfits.open(self.temp('test_append.fits'),
                             mode='append') as a:

                # The above change did not take effect since this was made
                # after the flush().
                assert_equal(a[1].data[1,1], 349)
                a.append(r[1])
            del a

            # When changes are made to an HDUList which was opened with
            # "update" mode, they will be written back to the original file
            # when a flush/close is called.
            os.rename(self.temp('test_append.fits'),
                      self.temp('test_update.fits'))

            with pyfits.open(self.temp('test_update.fits'),
                             mode='update') as u:

                # When the changes do not alter the size structures of the
                # original (or since last flush) HDUList, the changes are
                # written back "in place".
                assert_equal(u[0].header['rootname'], 'U2EQ0201T')
                u[0].header['rootname'] = 'abc'
                assert_equal(u[1].data[1,1], 349)
                u[1].data[1,1] = 99
                u.flush()

                # If the changes affect the size structure, e.g. adding or
                # deleting HDU(s), header was expanded or reduced beyond
                # existing number of blocks (2880 bytes in each block), or
                # change the data size, the HDUList is written to a temporary
                # file, the original file is deleted, and the temporary file is
                # renamed to the original file name and reopened in the update
                # mode.  To a user, these two kinds of updating writeback seem
                # to be the same, unless the optional argument in flush or
                # close is set to 1.
                del u[2]
                u.flush()

                # the write method in HDUList class writes the current HDUList,
                # with all changes made up to now, to a new file.  This method
                # works the same disregard the mode the HDUList was opened
                # with.
                u.append(r[3])
                u.writeto(self.temp('test_new.fits'))
            del u


        # Another useful new HDUList method is readall.  It will "touch" the
        # data parts in all HDUs, so even if the HDUList is closed, we can
        # still operate on the data.
        with pyfits.open(self.data('test0.fits')) as r:
            r.readall()
            assert_equal(r[1].data[1,1], 315)

        # create an HDU with data only
        data = np.ones((3,5), dtype=np.float32)
        hdu = pyfits.ImageHDU(data=data, name='SCI')
        assert_true((hdu.data ==
                     np.array([[ 1.,  1.,  1.,  1.,  1.],
                               [ 1.,  1.,  1.,  1.,  1.],
                               [ 1.,  1.,  1.,  1.,  1.]],
                              dtype=np.float32)).all())

        # create an HDU with header and data
        # notice that the header has the right NAXIS's since it is constructed
        # with ImageHDU
        hdu2 = pyfits.ImageHDU(header=r[1].header, data=np.array([1,2],
                               dtype='int32'))

        assert_equal(str(hdu2.header.ascard[1:5]),
            "BITPIX  =                   32 / array data type                                \n"
            "NAXIS   =                    1 / number of array dimensions                     \n"
            "NAXIS1  =                    2                                                  \n"
            "PCOUNT  =                    0 / number of parameters                           ")

    def test_memory_mapping(self):
        # memory mapping
        f1 = pyfits.open(self.data('test0.fits'), memmap=1)
        f1.close()

    def test_verification_on_output(self):
        # verification on output
        # make a defect HDUList first
        err_text = "HDUList's 0th element is not a primary HDU."
        with catch_warnings(record=True) as w:
            x = pyfits.ImageHDU()
            # HDUList can take a list or one single HDU
            hdu = pyfits.HDUList(x)
            with CaptureStdio():
                hdu.verify()
            assert_equal(len(w), 3)
            assert_true(err_text in str(w[1].message))

        fix_text = err_text + "  Fixed by inserting one as 0th HDU."
        with catch_warnings(record=True) as w:
            with CaptureStdio():
                hdu.writeto(self.temp('test_new2.fits'), 'fix')
            assert_equal(len(w), 3)
            assert_true(fix_text in str(w[1].message))

    def test_section(self):
        # section testing
        fs = pyfits.open(self.data('arange.fits'))
        assert_equal(fs[0].section[3,2,5], np.array([357]))
        assert_true((fs[0].section[3,2,:] ==
                     np.array([352, 353, 354, 355, 356, 357, 358, 359, 360,
                               361, 362])).all())
        assert_true((fs[0].section[3,2,4:] ==
                     np.array([356, 357, 358, 359, 360, 361, 362])).all())
        assert_true((fs[0].section[3,2,:8] ==
                     np.array([352, 353, 354, 355, 356, 357, 358, 359])).all())
        assert_true((fs[0].section[3,2,-8:8] ==
                     np.array([355, 356, 357, 358, 359])).all())
        assert_true((fs[0].section[3,2:5,:] ==
                     np.array([[352, 353, 354, 355, 356, 357, 358, 359,
                                360, 361, 362],
                               [363, 364, 365, 366, 367, 368, 369, 370,
                                371, 372, 373],
                               [374, 375, 376, 377, 378, 379, 380, 381,
                                382, 383, 384]])).all())

        assert_true((fs[0].section[3,:,:][:3,:3] ==
                     np.array([[330, 331, 332],
                               [341, 342, 343],
                               [352, 353, 354]])).all())

        dat = fs[0].data
        assert_true((fs[0].section[3,2:5,:8] == dat[3,2:5,:8]).all())
        assert_true((fs[0].section[3,2:5,3] == dat[3,2:5,3]).all())

        assert_true((fs[0].section[3:6,:,:][:3,:3,:3] ==
                     np.array([[[330, 331, 332],
                                [341, 342, 343],
                                [352, 353, 354]],
                               [[440, 441, 442],
                                [451, 452, 453],
                                [462, 463, 464]],
                               [[550, 551, 552],
                                [561, 562, 563],
                                [572, 573, 574]]])).all())

        assert_true((fs[0].section[:,:,:][:3,:2,:2] ==
                     np.array([[[  0,   1],
                                [ 11,  12]],
                               [[110, 111],
                                [121, 122]],
                               [[220, 221],
                                [231, 232]]])).all())

        assert_true((fs[0].section[:,2,:] == dat[:,2,:]).all())
        assert_true((fs[0].section[:,2:5,:] == dat[:,2:5,:]).all())
        assert_true((fs[0].section[3:6,3,:] == dat[3:6,3,:]).all())
        assert_true((fs[0].section[3:6,3:7,:] == dat[3:6,3:7,:]).all())

    def test_section_data_square(self):
        a = np.arange(4).reshape((2, 2))
        hdu = pyfits.PrimaryHDU(a)
        hdu.writeto(self.temp('test_new.fits'))

        hdul = pyfits.open(self.temp('test_new.fits'))
        d = hdul[0]
        dat = hdul[0].data
        assert_true((d.section[:,:] == dat[:,:]).all())
        assert_true((d.section[0,:] == dat[0,:]).all())
        assert_true((d.section[1,:] == dat[1,:]).all())
        assert_true((d.section[:,0] == dat[:,0]).all())
        assert_true((d.section[:,1] == dat[:,1]).all())
        assert_true((d.section[0,0] == dat[0,0]).all())
        assert_true((d.section[0,1] == dat[0,1]).all())
        assert_true((d.section[1,0] == dat[1,0]).all())
        assert_true((d.section[1,1] == dat[1,1]).all())
        assert_true((d.section[0:1,0:1] == dat[0:1,0:1]).all())
        assert_true((d.section[0:2,0:1] == dat[0:2,0:1]).all())
        assert_true((d.section[0:1,0:2] == dat[0:1,0:2]).all())
        assert_true((d.section[0:2,0:2] == dat[0:2,0:2]).all())

    def test_section_data_cube(self):
        a=np.arange(18).reshape((2,3,3))
        hdu = pyfits.PrimaryHDU(a)
        hdu.writeto(self.temp('test_new.fits'))

        hdul=pyfits.open(self.temp('test_new.fits'))
        d = hdul[0]
        dat = hdul[0].data
        assert_true((d.section[:,:,:] == dat[:,:,:]).all())
        assert_true((d.section[:,:] == dat[:,:]).all())
        assert_true((d.section[:] == dat[:]).all())
        assert_true((d.section[0,:,:] == dat[0,:,:]).all())
        assert_true((d.section[1,:,:] == dat[1,:,:]).all())
        assert_true((d.section[0,0,:] == dat[0,0,:]).all())
        assert_true((d.section[0,1,:] == dat[0,1,:]).all())
        assert_true((d.section[0,2,:] == dat[0,2,:]).all())
        assert_true((d.section[1,0,:] == dat[1,0,:]).all())
        assert_true((d.section[1,1,:] == dat[1,1,:]).all())
        assert_true((d.section[1,2,:] == dat[1,2,:]).all())
        assert_true((d.section[0,0,0] == dat[0,0,0]).all())
        assert_true((d.section[0,0,1] == dat[0,0,1]).all())
        assert_true((d.section[0,0,2] == dat[0,0,2]).all())
        assert_true((d.section[0,1,0] == dat[0,1,0]).all())
        assert_true((d.section[0,1,1] == dat[0,1,1]).all())
        assert_true((d.section[0,1,2] == dat[0,1,2]).all())
        assert_true((d.section[0,2,0] == dat[0,2,0]).all())
        assert_true((d.section[0,2,1] == dat[0,2,1]).all())
        assert_true((d.section[0,2,2] == dat[0,2,2]).all())
        assert_true((d.section[1,0,0] == dat[1,0,0]).all())
        assert_true((d.section[1,0,1] == dat[1,0,1]).all())
        assert_true((d.section[1,0,2] == dat[1,0,2]).all())
        assert_true((d.section[1,1,0] == dat[1,1,0]).all())
        assert_true((d.section[1,1,1] == dat[1,1,1]).all())
        assert_true((d.section[1,1,2] == dat[1,1,2]).all())
        assert_true((d.section[1,2,0] == dat[1,2,0]).all())
        assert_true((d.section[1,2,1] == dat[1,2,1]).all())
        assert_true((d.section[1,2,2] == dat[1,2,2]).all())
        assert_true((d.section[:,0,0] == dat[:,0,0]).all())
        assert_true((d.section[:,0,1] == dat[:,0,1]).all())
        assert_true((d.section[:,0,2] == dat[:,0,2]).all())
        assert_true((d.section[:,1,0] == dat[:,1,0]).all())
        assert_true((d.section[:,1,1] == dat[:,1,1]).all())
        assert_true((d.section[:,1,2] == dat[:,1,2]).all())
        assert_true((d.section[:,2,0] == dat[:,2,0]).all())
        assert_true((d.section[:,2,1] == dat[:,2,1]).all())
        assert_true((d.section[:,2,2] == dat[:,2,2]).all())
        assert_true((d.section[0,:,0] == dat[0,:,0]).all())
        assert_true((d.section[0,:,1] == dat[0,:,1]).all())
        assert_true((d.section[0,:,2] == dat[0,:,2]).all())
        assert_true((d.section[1,:,0] == dat[1,:,0]).all())
        assert_true((d.section[1,:,1] == dat[1,:,1]).all())
        assert_true((d.section[1,:,2] == dat[1,:,2]).all())
        assert_true((d.section[:,:,0] == dat[:,:,0]).all())
        assert_true((d.section[:,:,1] == dat[:,:,1]).all())
        assert_true((d.section[:,:,2] == dat[:,:,2]).all())
        assert_true((d.section[:,0,:] == dat[:,0,:]).all())
        assert_true((d.section[:,1,:] == dat[:,1,:]).all())
        assert_true((d.section[:,2,:] == dat[:,2,:]).all())

        assert_true((d.section[:,:,0:1] == dat[:,:,0:1]).all())
        assert_true((d.section[:,:,0:2] == dat[:,:,0:2]).all())
        assert_true((d.section[:,:,0:3] == dat[:,:,0:3]).all())
        assert_true((d.section[:,:,1:2] == dat[:,:,1:2]).all())
        assert_true((d.section[:,:,1:3] == dat[:,:,1:3]).all())
        assert_true((d.section[:,:,2:3] == dat[:,:,2:3]).all())
        assert_true((d.section[0:1,0:1,0:1] == dat[0:1,0:1,0:1]).all())
        assert_true((d.section[0:1,0:1,0:2] == dat[0:1,0:1,0:2]).all())
        assert_true((d.section[0:1,0:1,0:3] == dat[0:1,0:1,0:3]).all())
        assert_true((d.section[0:1,0:1,1:2] == dat[0:1,0:1,1:2]).all())
        assert_true((d.section[0:1,0:1,1:3] == dat[0:1,0:1,1:3]).all())
        assert_true((d.section[0:1,0:1,2:3] == dat[0:1,0:1,2:3]).all())
        assert_true((d.section[0:1,0:2,0:1] == dat[0:1,0:2,0:1]).all())
        assert_true((d.section[0:1,0:2,0:2] == dat[0:1,0:2,0:2]).all())
        assert_true((d.section[0:1,0:2,0:3] == dat[0:1,0:2,0:3]).all())
        assert_true((d.section[0:1,0:2,1:2] == dat[0:1,0:2,1:2]).all())
        assert_true((d.section[0:1,0:2,1:3] == dat[0:1,0:2,1:3]).all())
        assert_true((d.section[0:1,0:2,2:3] == dat[0:1,0:2,2:3]).all())
        assert_true((d.section[0:1,0:3,0:1] == dat[0:1,0:3,0:1]).all())
        assert_true((d.section[0:1,0:3,0:2] == dat[0:1,0:3,0:2]).all())
        assert_true((d.section[0:1,0:3,0:3] == dat[0:1,0:3,0:3]).all())
        assert_true((d.section[0:1,0:3,1:2] == dat[0:1,0:3,1:2]).all())
        assert_true((d.section[0:1,0:3,1:3] == dat[0:1,0:3,1:3]).all())
        assert_true((d.section[0:1,0:3,2:3] == dat[0:1,0:3,2:3]).all())
        assert_true((d.section[0:1,1:2,0:1] == dat[0:1,1:2,0:1]).all())
        assert_true((d.section[0:1,1:2,0:2] == dat[0:1,1:2,0:2]).all())
        assert_true((d.section[0:1,1:2,0:3] == dat[0:1,1:2,0:3]).all())
        assert_true((d.section[0:1,1:2,1:2] == dat[0:1,1:2,1:2]).all())
        assert_true((d.section[0:1,1:2,1:3] == dat[0:1,1:2,1:3]).all())
        assert_true((d.section[0:1,1:2,2:3] == dat[0:1,1:2,2:3]).all())
        assert_true((d.section[0:1,1:3,0:1] == dat[0:1,1:3,0:1]).all())
        assert_true((d.section[0:1,1:3,0:2] == dat[0:1,1:3,0:2]).all())
        assert_true((d.section[0:1,1:3,0:3] == dat[0:1,1:3,0:3]).all())
        assert_true((d.section[0:1,1:3,1:2] == dat[0:1,1:3,1:2]).all())
        assert_true((d.section[0:1,1:3,1:3] == dat[0:1,1:3,1:3]).all())
        assert_true((d.section[0:1,1:3,2:3] == dat[0:1,1:3,2:3]).all())
        assert_true((d.section[1:2,0:1,0:1] == dat[1:2,0:1,0:1]).all())
        assert_true((d.section[1:2,0:1,0:2] == dat[1:2,0:1,0:2]).all())
        assert_true((d.section[1:2,0:1,0:3] == dat[1:2,0:1,0:3]).all())
        assert_true((d.section[1:2,0:1,1:2] == dat[1:2,0:1,1:2]).all())
        assert_true((d.section[1:2,0:1,1:3] == dat[1:2,0:1,1:3]).all())
        assert_true((d.section[1:2,0:1,2:3] == dat[1:2,0:1,2:3]).all())
        assert_true((d.section[1:2,0:2,0:1] == dat[1:2,0:2,0:1]).all())
        assert_true((d.section[1:2,0:2,0:2] == dat[1:2,0:2,0:2]).all())
        assert_true((d.section[1:2,0:2,0:3] == dat[1:2,0:2,0:3]).all())
        assert_true((d.section[1:2,0:2,1:2] == dat[1:2,0:2,1:2]).all())
        assert_true((d.section[1:2,0:2,1:3] == dat[1:2,0:2,1:3]).all())
        assert_true((d.section[1:2,0:2,2:3] == dat[1:2,0:2,2:3]).all())
        assert_true((d.section[1:2,0:3,0:1] == dat[1:2,0:3,0:1]).all())
        assert_true((d.section[1:2,0:3,0:2] == dat[1:2,0:3,0:2]).all())
        assert_true((d.section[1:2,0:3,0:3] == dat[1:2,0:3,0:3]).all())
        assert_true((d.section[1:2,0:3,1:2] == dat[1:2,0:3,1:2]).all())
        assert_true((d.section[1:2,0:3,1:3] == dat[1:2,0:3,1:3]).all())
        assert_true((d.section[1:2,0:3,2:3] == dat[1:2,0:3,2:3]).all())
        assert_true((d.section[1:2,1:2,0:1] == dat[1:2,1:2,0:1]).all())
        assert_true((d.section[1:2,1:2,0:2] == dat[1:2,1:2,0:2]).all())
        assert_true((d.section[1:2,1:2,0:3] == dat[1:2,1:2,0:3]).all())
        assert_true((d.section[1:2,1:2,1:2] == dat[1:2,1:2,1:2]).all())
        assert_true((d.section[1:2,1:2,1:3] == dat[1:2,1:2,1:3]).all())
        assert_true((d.section[1:2,1:2,2:3] == dat[1:2,1:2,2:3]).all())
        assert_true((d.section[1:2,1:3,0:1] == dat[1:2,1:3,0:1]).all())
        assert_true((d.section[1:2,1:3,0:2] == dat[1:2,1:3,0:2]).all())
        assert_true((d.section[1:2,1:3,0:3] == dat[1:2,1:3,0:3]).all())
        assert_true((d.section[1:2,1:3,1:2] == dat[1:2,1:3,1:2]).all())
        assert_true((d.section[1:2,1:3,1:3] == dat[1:2,1:3,1:3]).all())
        assert_true((d.section[1:2,1:3,2:3] == dat[1:2,1:3,2:3]).all())

    def test_section_data_four(self):
        a = np.arange(256).reshape((4, 4, 4, 4))
        hdu = pyfits.PrimaryHDU(a)
        hdu.writeto(self.temp('test_new.fits'))

        hdul = pyfits.open(self.temp('test_new.fits'))
        d = hdul[0]
        dat = hdul[0].data
        assert_true((d.section[:,:,:,:] == dat[:,:,:,:]).all())
        assert_true((d.section[:,:,:] == dat[:,:,:]).all())
        assert_true((d.section[:,:] == dat[:,:]).all())
        assert_true((d.section[:] == dat[:]).all())
        assert_true((d.section[0,:,:,:] == dat[0,:,:,:]).all())
        assert_true((d.section[0,:,0,:] == dat[0,:,0,:]).all())
        assert_true((d.section[:,:,0,:] == dat[:,:,0,:]).all())
        assert_true((d.section[:,1,0,:] == dat[:,1,0,:]).all())
        assert_true((d.section[:,:,:,1] == dat[:,:,:,1]).all())

    def test_section_data_scaled(self):
        """
        Regression test for #143.  This is like test_section_data_square but
        uses a file containing scaled image data, to test that sections can
        work correctly with scaled data.
        """

        hdul = pyfits.open(self.data('scale.fits'))
        d = hdul[0]
        dat = hdul[0].data
        assert_true((d.section[:,:] == dat[:,:]).all())
        assert_true((d.section[0,:] == dat[0,:]).all())
        assert_true((d.section[1,:] == dat[1,:]).all())
        assert_true((d.section[:,0] == dat[:,0]).all())
        assert_true((d.section[:,1] == dat[:,1]).all())
        assert_true((d.section[0,0] == dat[0,0]).all())
        assert_true((d.section[0,1] == dat[0,1]).all())
        assert_true((d.section[1,0] == dat[1,0]).all())
        assert_true((d.section[1,1] == dat[1,1]).all())
        assert_true((d.section[0:1,0:1] == dat[0:1,0:1]).all())
        assert_true((d.section[0:2,0:1] == dat[0:2,0:1]).all())
        assert_true((d.section[0:1,0:2] == dat[0:1,0:2]).all())
        assert_true((d.section[0:2,0:2] == dat[0:2,0:2]).all())

        # Test without having accessed the full data first
        hdul = pyfits.open(self.data('scale.fits'))
        d = hdul[0]
        assert_true((d.section[:,:] == dat[:,:]).all())
        assert_true((d.section[0,:] == dat[0,:]).all())
        assert_true((d.section[1,:] == dat[1,:]).all())
        assert_true((d.section[:,0] == dat[:,0]).all())
        assert_true((d.section[:,1] == dat[:,1]).all())
        assert_true((d.section[0,0] == dat[0,0]).all())
        assert_true((d.section[0,1] == dat[0,1]).all())
        assert_true((d.section[1,0] == dat[1,0]).all())
        assert_true((d.section[1,1] == dat[1,1]).all())
        assert_true((d.section[0:1,0:1] == dat[0:1,0:1]).all())
        assert_true((d.section[0:2,0:1] == dat[0:2,0:1]).all())
        assert_true((d.section[0:1,0:2] == dat[0:1,0:2]).all())
        assert_true((d.section[0:2,0:2] == dat[0:2,0:2]).all())
        assert_false(d._data_loaded)

    def test_comp_image(self):
        argslist = [
            (np.zeros((2, 10, 10), dtype=np.float32), 'RICE_1', 16),
            (np.zeros((2, 10, 10), dtype=np.float32), 'GZIP_1', -0.01),
            (np.zeros((100, 100)) + 1, 'HCOMPRESS_1', 16)
        ]

        for byte_order in ('<', '>'):
            for args in argslist:
                yield (self._test_comp_image,) + args + (byte_order,)

    def _test_comp_image(self, data, compression_type, quantize_level,
                         byte_order):
        data = data.newbyteorder(byte_order)
        primary_hdu = pyfits.PrimaryHDU()
        ofd = pyfits.HDUList(primary_hdu)
        chdu = pyfits.CompImageHDU(data, name='SCI',
                                   compressionType=compression_type,
                                   quantizeLevel=quantize_level)
        ofd.append(chdu)
        ofd.writeto(self.temp('test_new.fits'), clobber=True)
        ofd.close()
        with pyfits.open(self.temp('test_new.fits')) as fd:
            assert_true((fd[1].data == data).all())
            assert_equal(fd[1].header['NAXIS'], chdu.header['NAXIS'])
            assert_equal(fd[1].header['NAXIS1'], chdu.header['NAXIS1'])
            assert_equal(fd[1].header['NAXIS2'], chdu.header['NAXIS2'])
            assert_equal(fd[1].header['BITPIX'], chdu.header['BITPIX'])

    def test_comp_image_hcompression_1_invalid_data(self):
        """
        Tests compression with the HCOMPRESS_1 algorithm with data that is
        not 2D and has a non-2D tile size.
        """

        assert_raises(ValueError, pyfits.CompImageHDU,
                      np.zeros((2, 10, 10), dtype=np.float32), name='SCI',
                      compressionType='HCOMPRESS_1', quantizeLevel=16,
                      tileSize=[2, 10, 10])

    def test_comp_image_hcompress_image_stack(self):
        """
        Regression test for #171.

        Tests that data containing more than two dimensions can be
        compressed with HCOMPRESS_1 so long as the user-supplied tile size can
        be flattened to two dimensions.
        """

        cube = np.arange(300, dtype=np.float32).reshape((3, 10, 10))
        hdu = pyfits.CompImageHDU(data=cube, name='SCI',
                                  compressionType='HCOMPRESS_1',
                                  quantizeLevel=16, tileSize=[5, 5, 1])
        hdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits')) as hdul:
            assert_true((hdul['SCI'].data == cube).all())

    def test_disable_image_compression(self):
        with catch_warnings():
            # No warnings should be displayed in this case
            warnings.simplefilter('error')
            with pyfits.open(self.data('comp.fits'),
                             disable_image_compression=True) as hdul:
                # The compressed image HDU should show up as a BinTableHDU, but
                # *not* a CompImageHDU
                assert_true(isinstance(hdul[1], pyfits.BinTableHDU))
                assert_false(isinstance(hdul[1], pyfits.CompImageHDU))

        with pyfits.open(self.data('comp.fits')) as hdul:
            assert_true(isinstance(hdul[1], pyfits.CompImageHDU))

    def test_open_comp_image_in_update_mode(self):
        """
        Regression test for #167.

        Similar to test_open_scaled_in_update_mode(), but specifically for
        compressed images.
        """

        # Copy the original file before making any possible changes to it
        self.copy_file('comp.fits')
        mtime = os.stat(self.temp('comp.fits')).st_mtime

        time.sleep(1)

        pyfits.open(self.temp('comp.fits'), mode='update').close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert_equal(mtime, os.stat(self.temp('comp.fits')).st_mtime)

    def test_do_not_scale_image_data(self):
        hdul = pyfits.open(self.data('scale.fits'),
                           do_not_scale_image_data=True)
        assert_equal(hdul[0].data.dtype, np.dtype('>i2'))
        hdul = pyfits.open(self.data('scale.fits'))
        assert_equal(hdul[0].data.dtype, np.dtype('float32'))

    def test_append_uint_data(self):
        """Regression test for #56 (BZERO and BSCALE added in the wrong location
        when appending scaled data)
        """

        pyfits.writeto(self.temp('test_new.fits'), data=np.array([],
                       dtype='uint8'))
        d = np.zeros([100, 100]).astype('uint16')
        pyfits.append(self.temp('test_new.fits'), data=d)
        f = pyfits.open(self.temp('test_new.fits'), uint=True)
        assert_equal(f[1].data.dtype, 'uint16')

    def test_blanks(self):
        """Test image data with blank spots in it (which should show up as
        NaNs in the data array.
        """

        arr = np.zeros((10, 10), dtype=np.int32)
        # One row will be blanks
        arr[1] = 999
        hdu = pyfits.ImageHDU(data=arr)
        hdu.header['BLANK'] = 999
        hdu.writeto(self.temp('test_new.fits'))

        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true(np.isnan(hdul[1].data[1]).all())

    def test_bzero_with_floats(self):
        """Test use of the BZERO keyword in an image HDU containing float
        data.
        """

        arr = np.zeros((10, 10)) - 1
        hdu = pyfits.ImageHDU(data=arr)
        hdu.header['BZERO'] = 1.0
        hdu.writeto(self.temp('test_new.fits'))

        hdul = pyfits.open(self.temp('test_new.fits'))
        arr += 1
        assert_true((hdul[1].data == arr).all())

    def test_rewriting_large_scaled_image(self):
        """Regression test for #84 and #101."""

        hdul = pyfits.open(self.data('fixed-1890.fits'))
        orig_data = hdul[0].data
        with ignore_warnings():
            hdul.writeto(self.temp('test_new.fits'), clobber=True)
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[0].data == orig_data).all())
        hdul.close()

        # Just as before, but this time don't touch hdul[0].data before writing
        # back out--this is the case that failed in #84
        hdul = pyfits.open(self.data('fixed-1890.fits'))
        with ignore_warnings():
            hdul.writeto(self.temp('test_new.fits'), clobber=True)
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[0].data == orig_data).all())
        hdul.close()

        # Test opening/closing/reopening a scaled file in update mode
        hdul = pyfits.open(self.data('fixed-1890.fits'),
                           do_not_scale_image_data=True)
        hdul.writeto(self.temp('test_new.fits'), clobber=True,
                     output_verify='silentfix')
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        orig_data = hdul[0].data
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'), mode='update')
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[0].data == orig_data).all())
        hdul = pyfits.open(self.temp('test_new.fits'))
        hdul.close()

    def test_rewriting_large_scaled_image_compressed(self):
        """
        Regression test for #88 1.

        Identical to test_rewriting_large_scaled_image() but with a compressed
        image.
        """

        with pyfits.open(self.data('fixed-1890.fits'),
                         do_not_scale_image_data=True) as hdul:
            chdu = pyfits.CompImageHDU(data=hdul[0].data,
                                       header=hdul[0].header)
            chdu.writeto(self.temp('fixed-1890-z.fits'))

        hdul = pyfits.open(self.temp('fixed-1890-z.fits'))
        orig_data = hdul[1].data
        with ignore_warnings():
            hdul.writeto(self.temp('test_new.fits'), clobber=True)
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[1].data == orig_data).all())
        hdul.close()

        # Just as before, but this time don't touch hdul[0].data before writing
        # back out--this is the case that failed in #84
        hdul = pyfits.open(self.temp('fixed-1890-z.fits'))
        with ignore_warnings():
            hdul.writeto(self.temp('test_new.fits'), clobber=True)
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[1].data == orig_data).all())
        hdul.close()

        # Test opening/closing/reopening a scaled file in update mode
        hdul = pyfits.open(self.temp('fixed-1890-z.fits'),
                           do_not_scale_image_data=True)
        hdul.writeto(self.temp('test_new.fits'), clobber=True,
                     output_verify='silentfix')
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        orig_data = hdul[1].data
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'), mode='update')
        hdul.close()
        hdul = pyfits.open(self.temp('test_new.fits'))
        assert_true((hdul[1].data == orig_data).all())
        hdul = pyfits.open(self.temp('test_new.fits'))
        hdul.close()

    def test_image_update_header(self):
        """
        Regression test for #105.

        Replacing the original header to an image HDU and saving should update
        the NAXISn keywords appropriately and save the image data correctly.
        """

        # Copy the original file before saving to it
        self.copy_file('test0.fits')
        with pyfits.open(self.temp('test0.fits'), mode='update') as hdul:
            orig_data = hdul[1].data.copy()
            hdr_copy = hdul[1].header.copy()
            del hdr_copy['NAXIS*']
            hdul[1].header = hdr_copy

        with pyfits.open(self.temp('test0.fits')) as hdul:
            assert_true((orig_data == hdul[1].data).all())

    def test_open_scaled_in_update_mode(self):
        """
        Regression test for #119 (Don't update scaled image data if the data is
        not read)

        This ensures that merely opening and closing a file containing scaled
        image data does not cause any change to the data (or the header).
        Changes should only occur if the data is accessed.
        """

        # Copy the original file before making any possible changes to it
        self.copy_file('scale.fits')
        mtime = os.stat(self.temp('scale.fits')).st_mtime

        time.sleep(1)

        pyfits.open(self.temp('scale.fits'), mode='update').close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert_equal(mtime, os.stat(self.temp('scale.fits')).st_mtime)

        # Insert a slight delay to ensure the mtime does change when the file
        # is changed
        time.sleep(1)

        hdul = pyfits.open(self.temp('scale.fits'), 'update')
        hdul[0].data
        hdul.close()

        # Now the file should be updated with the rescaled data
        assert_not_equal(mtime, os.stat(self.temp('scale.fits')).st_mtime)
        hdul = pyfits.open(self.temp('scale.fits'), mode='update')
        assert_equal(hdul[0].data.dtype, np.dtype('>f4'))
        assert_equal(hdul[0].header['BITPIX'], -32)
        assert_true('BZERO' not in hdul[0].header)
        assert_true('BSCALE' not in hdul[0].header)

        # Try reshaping the data, then closing and reopening the file; let's
        # see if all the changes are preseved properly
        hdul[0].data.shape = (42, 10)
        hdul.close()

        hdul = pyfits.open(self.temp('scale.fits'))
        assert_equal(hdul[0].shape, (42, 10))
        assert_equal(hdul[0].data.dtype, np.dtype('>f4'))
        assert_equal(hdul[0].header['BITPIX'], -32)
        assert_true('BZERO' not in hdul[0].header)
        assert_true('BSCALE' not in hdul[0].header)

    def test_open_scaled_in_update_mode_compressed(self):
        """
        Regression test for #88 2.

        Identical to test_open_scaled_in_update_mode() but with a compressed
        version of the scaled image.
        """

        # Copy+compress the original file before making any possible changes to
        # it
        with pyfits.open(self.data('scale.fits'),
                         do_not_scale_image_data=True) as hdul:
            chdu = pyfits.CompImageHDU(data=hdul[0].data,
                                       header=hdul[0].header)
            chdu.writeto(self.temp('scale.fits'))
        mtime = os.stat(self.temp('scale.fits')).st_mtime

        time.sleep(1)

        pyfits.open(self.temp('scale.fits'), mode='update').close()

        # Ensure that no changes were made to the file merely by immediately
        # opening and closing it.
        assert_equal(mtime, os.stat(self.temp('scale.fits')).st_mtime)

        # Insert a slight delay to ensure the mtime does change when the file
        # is changed
        time.sleep(1)

        hdul = pyfits.open(self.temp('scale.fits'), 'update')
        hdul[1].data
        hdul.close()

        # Now the file should be updated with the rescaled data
        assert_not_equal(mtime, os.stat(self.temp('scale.fits')).st_mtime)
        hdul = pyfits.open(self.temp('scale.fits'), mode='update')
        assert_equal(hdul[1].data.dtype, np.dtype('float32'))
        assert_equal(hdul[1].header['BITPIX'], -32)
        assert_true('BZERO' not in hdul[1].header)
        assert_true('BSCALE' not in hdul[1].header)

        # Try reshaping the data, then closing and reopening the file; let's
        # see if all the changes are preseved properly
        hdul[1].data.shape = (42, 10)
        hdul.close()

        hdul = pyfits.open(self.temp('scale.fits'))
        assert_equal(hdul[1].shape, (42, 10))
        assert_equal(hdul[1].data.dtype, np.dtype('float32'))
        assert_equal(hdul[1].header['BITPIX'], -32)
        assert_true('BZERO' not in hdul[1].header)
        assert_true('BSCALE' not in hdul[1].header)

    def test_scale_back(self):
        """A simple test for #120--the scale_back feature for image HDUs."""

        self.copy_file('scale.fits')
        with pyfits.open(self.temp('scale.fits'), mode='update',
                         scale_back=True) as hdul:
            orig_bitpix = hdul[0].header['BITPIX']
            orig_bzero = hdul[0].header['BZERO']
            orig_bscale = hdul[0].header['BSCALE']
            orig_data = hdul[0].data.copy()
            hdul[0].data[0] = 0

        with pyfits.open(self.temp('scale.fits'),
                         do_not_scale_image_data=True) as hdul:
            assert_equal(hdul[0].header['BITPIX'], orig_bitpix)
            assert_equal(hdul[0].header['BZERO'], orig_bzero)
            assert_equal(hdul[0].header['BSCALE'], orig_bscale)

            zero_point = int(math.floor(-orig_bzero / orig_bscale))
            assert_true((hdul[0].data[0] == zero_point).all())

        with pyfits.open(self.temp('scale.fits')) as hdul:
            assert_true((hdul[0].data[1:] == orig_data[1:]).all())

    def test_scale_back_compressed(self):
        """
        Regression test for #88 3.

        Identical to test_scale_back() but uses a compressed image.
        """

        # Create a compressed version of the scaled image
        with pyfits.open(self.data('scale.fits'),
                         do_not_scale_image_data=True) as hdul:
            chdu = pyfits.CompImageHDU(data=hdul[0].data,
                                       header=hdul[0].header)
            chdu.writeto(self.temp('scale.fits'))

        with pyfits.open(self.temp('scale.fits'), mode='update',
                         scale_back=True) as hdul:
            orig_bitpix = hdul[1].header['BITPIX']
            orig_bzero = hdul[1].header['BZERO']
            orig_bscale = hdul[1].header['BSCALE']
            orig_data = hdul[1].data.copy()
            hdul[1].data[0] = 0

        with pyfits.open(self.temp('scale.fits'),
                         do_not_scale_image_data=True) as hdul:
            assert_equal(hdul[1].header['BITPIX'], orig_bitpix)
            assert_equal(hdul[1].header['BZERO'], orig_bzero)
            assert_equal(hdul[1].header['BSCALE'], orig_bscale)

            zero_point = int(math.floor(-orig_bzero / orig_bscale))
            assert_true((hdul[1].data[0] == zero_point).all())

        with pyfits.open(self.temp('scale.fits')) as hdul:
            assert_true((hdul[1].data[1:] == orig_data[1:]).all())
            # Extra test to ensure that after everything the data is still the
            # same as in the original uncompressed version of the image
            with pyfits.open(self.data('scale.fits')) as hdul2:
                # Recall we made the same modification to the data in hdul
                # above
                hdul2[0].data[0] = 0
                assert_true((hdul[1].data == hdul2[0].data).all())

    def test_compression_column_tforms(self):
        """Regression test for #199."""

        # Some interestingly tiled data so that some of it is quantized and
        # some of it ends up just getting gzip-compressed
        data2 = ((np.arange(1, 8, dtype=np.float32) * 10)[:, np.newaxis] +
                np.arange(1, 7))
        np.random.seed(1337)
        data1 = np.random.uniform(size=(6 * 4, 7 * 4))
        data1[:data2.shape[0], :data2.shape[1]] = data2
        chdu = pyfits.CompImageHDU(data1, compressionType='RICE_1',
                                   tileSize=(6, 7))
        chdu.writeto(self.temp('test.fits'))

        with pyfits.open(self.temp('test.fits'),
                         disable_image_compression=True) as h:
            assert_equal(h[1].header['TFORM1'], '1PB(51)')
            assert_equal(h[1].header['TFORM2'], '1PD(2)')

        with pyfits.open(self.temp('test.fits')) as h:
            assert_allclose(h[1].data, data1, rtol=0.1, atol=0.1)
